from collections import defaultdict


class MDSResponse:
    def __init__(self, response_proto, metadata=defaultdict(str)):
        """
        A wrapper of response protos from metadata server that adds a method to access response headers
        This class is needed since we cannot add an attribute to Python proto Message classes due to their designs which is documented here:
            https://chromium.googlesource.com/chromium/src/third_party/+/master/protobuf/python/google/protobuf/internal/python_message.py

        :param response_proto (:Message): deserialized response proto returned from metadata server
        :param metadata (:defaultdict(str)): response header returned from metadata server
            If metadata is returned from go proxy (direct http), metadata has type Dict[str, List[str]].
            If metadata is returned from go proxy (grpc gateway), metadata has type Dict[str, str] whose value is first str in List[str].
            If metadata is returned from metadata server (direct grpc), metadata has type Dict[str, str].
            Currently, only _CLIENT_VERSION_INFO_RESPONSE_HEADER ad _CLIENT_UPDATE_VERSION_RESPONSE_HEADER metadata is used in the response,
             whose values have str type.
            The default values of keys that don't exist are empty strings in any of the 3 cases.
            It's possible to have other values for response. Some examples could be found here:
                https://grpc.github.io/grpc/python/glossary.html#term-metadata
                and here: https://www.tabnine.com/code/java/classes/io.grpc.Metadata$Key
                so the returned value may change in the future
        """
        self.response_proto = response_proto
        self.metadata = metadata

    def __getattr__(self, name):
        if isinstance(self.response_proto, dict):
            return self.response_proto[name]
        return getattr(self.response_proto, name)

    def _headers(self):
        return self.metadata



from datetime import datetime
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from urllib.parse import urljoin

from tecton._internals import errors
from tecton_core import conf
from tecton_core import http


class IngestionClient:
    def __init__(self):
        domain = self._get_domain()
        self.ingestion_url: str = urljoin(domain + "/", "ingest")

    @staticmethod
    def _get_domain():
        test_domain = conf.get_or_none("INGESTION_SERVICE")
        if test_domain:
            return test_domain
        preview_domain = conf.get_or_raise("FEATURE_SERVICE")
        preview_domain = preview_domain.replace("https://", "https://preview.")
        if preview_domain.endswith("/"):
            preview_domain = preview_domain[:-1]
        if preview_domain.endswith("/api"):
            preview_domain = preview_domain[:-4]
        return preview_domain

    def ingest(
        self,
        workspace_name: str,
        push_source_name: str,
        ingestion_records: Sequence[Dict[str, Any]],
        dry_run: bool = False,
    ) -> Tuple[int, str, Dict[str, Any]]:
        http_request = {
            "workspace_name": workspace_name,
            "dry_run": dry_run,
            "records": {
                push_source_name: [
                    {
                        "record": self._serialize_record(ingestion_record),
                    }
                    for ingestion_record in ingestion_records
                ]
            },
        }
        response = http.session().post(self.ingestion_url, json=http_request, headers=self._prepare_headers())

        if response.status_code != 200:
            error_message = f"Failed to ingest records: {response.reason}"
            return response.status_code, error_message, None

        try:
            response_data = response.json()
        except ValueError:
            error_message = "Response content is not valid JSON"
            return response.status_code, error_message, None
        return response.status_code, response.reason, response_data

    @staticmethod
    def _prepare_headers() -> Dict[str, str]:
        token = conf.get_or_none("TECTON_API_KEY")
        if not token:
            raise errors.FS_API_KEY_MISSING

        return {
            "authorization": f"Tecton-key {token}",
            "Content-Type": "application/json",
        }

    @staticmethod
    def _serialize_record(record: Dict[str, Any]) -> Dict[str, Any]:
        serialized_record = {}
        for k, v in record.items():
            if isinstance(v, datetime):
                serialized_record[k] = v.isoformat()
            else:
                serialized_record[k] = v
        return serialized_record



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

from pyspark.rdd import PythonEvalType


class PandasMapOpsMixin(object):
    """
    Min-in for pandas map operations. Currently, only :class:`DataFrame`
    can use this class.
    """

    def mapInPandas(self, func, schema):
        """
        Maps an iterator of batches in the current :class:`DataFrame` using a Python native
        function that takes and outputs a pandas DataFrame, and returns the result as a
        :class:`DataFrame`.

        The function should take an iterator of `pandas.DataFrame`\\s and return
        another iterator of `pandas.DataFrame`\\s. All columns are passed
        together as an iterator of `pandas.DataFrame`\\s to the function and the
        returned iterator of `pandas.DataFrame`\\s are combined as a :class:`DataFrame`.
        Each `pandas.DataFrame` size can be controlled by
        `spark.sql.execution.arrow.maxRecordsPerBatch`.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        func : function
            a Python native function that takes an iterator of `pandas.DataFrame`\\s, and
            outputs an iterator of `pandas.DataFrame`\\s.
        schema : :class:`pyspark.sql.types.DataType` or str
            the return type of the `func` in PySpark. The value can be either a
            :class:`pyspark.sql.types.DataType` object or a DDL-formatted type string.

        Examples
        --------
        >>> from pyspark.sql.functions import pandas_udf
        >>> df = spark.createDataFrame([(1, 21), (2, 30)], ("id", "age"))
        >>> def filter_func(iterator):
        ...     for pdf in iterator:
        ...         yield pdf[pdf.id == 1]
        >>> df.mapInPandas(filter_func, df.schema).show()  # doctest: +SKIP
        +---+---+
        | id|age|
        +---+---+
        |  1| 21|
        +---+---+

        Notes
        -----
        This API is experimental

        See Also
        --------
        pyspark.sql.functions.pandas_udf
        """
        from pyspark.sql import DataFrame
        from pyspark.sql.pandas.functions import pandas_udf

        assert isinstance(self, DataFrame)

        udf = pandas_udf(
            func, returnType=schema, functionType=PythonEvalType.SQL_MAP_PANDAS_ITER_UDF)
        udf_column = udf(*[self[col] for col in self.columns])
        jdf = self._jdf.mapInPandas(udf_column._jc.expr())
        return DataFrame(jdf, self.sql_ctx)


def _test():
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.sql.pandas.map_ops
    globs = pyspark.sql.pandas.map_ops.__dict__.copy()
    spark = SparkSession.builder\
        .master("local[4]")\
        .appName("sql.pandas.map_ops tests")\
        .getOrCreate()
    globs['spark'] = spark
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.pandas.map_ops, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



from types import MappingProxyType
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pendulum
from typeguard import typechecked

from tecton_core import conf
from tecton_core import errors
from tecton_core import function_deserialization
from tecton_core import time_utils
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import data_source__client_pb2 as data_source__args_pb2
from tecton_proto.args import data_source_config__client_pb2 as data_source_config_pb2
from tecton_proto.args import virtual_data_source__client_pb2 as virtual_data_source__args_pb2
from tecton_proto.common import data_source_type__client_pb2 as data_source_type_pb2
from tecton_proto.common import schema_container__client_pb2 as schema_container_pb2
from tecton_proto.common import secret__client_pb2 as secret_pb2
from tecton_proto.common import spark_schema__client_pb2 as spark_schema_pb2
from tecton_proto.data import batch_data_source__client_pb2 as batch_data_source__data_pb2
from tecton_proto.data import stream_data_source__client_pb2 as stream_data_source__data_pb2
from tecton_proto.data import virtual_data_source__client_pb2 as virtual_data_source__data_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = [
    "DataSourceSpec",
    "DataSourceSpecArgsSupplement",
    "BatchSourceSpec",
    "HiveSourceSpec",
    "UnitySourceSpec",
    "FileSourceSpec",
    "SparkBatchSourceSpec",
    "PandasBatchSourceSpec",
    "PushTableSourceSpec",
    "RedshiftSourceSpec",
    "SnowflakeSourceSpec",
    "BigquerySourceSpec",
    "DatetimePartitionColumnSpec",
    "StreamSourceSpec",
    "KinesisSourceSpec",
    "KafkaSourceSpec",
    "SparkStreamSourceSpec",
]


@utils.frozen_strict
class DataSourceSpec(tecton_object_spec.TectonObjectSpec):
    batch_source: Optional["BatchSourceSpec"]
    stream_source: Optional["StreamSourceSpec"]
    schema: Optional[schema_container_pb2.SchemaContainer]
    type: data_source_type_pb2.DataSourceType.ValueType

    options: Mapping[str, str]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: virtual_data_source__data_pb2.VirtualDataSource, deserialize_funcs_to_main: bool = False
    ) -> "DataSourceSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.virtual_data_source_id, proto.fco_metadata
            ),
            batch_source=create_batch_source_from_data_proto(proto.batch_data_source, deserialize_funcs_to_main),
            stream_source=create_stream_source_from_data_proto(proto.stream_data_source, deserialize_funcs_to_main),
            type=utils.get_field_or_none(proto, "data_source_type"),
            schema=utils.get_field_or_none(proto, "schema"),
            validation_args=validator_pb2.FcoValidationArgs(virtual_data_source=proto.validation_args),
            options=MappingProxyType(proto.options),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls,
        proto: virtual_data_source__args_pb2.VirtualDataSourceArgs,
        supplement: "DataSourceSpecArgsSupplement",
    ) -> "DataSourceSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(
                proto.virtual_data_source_id, proto.info
            ),
            batch_source=create_batch_source_from_args_proto(proto, supplement),
            stream_source=create_stream_source_from_args_proto(proto, supplement),
            schema=utils.get_field_or_none(proto, "schema"),
            type=utils.get_field_or_none(proto, "type"),
            validation_args=None,
            options=MappingProxyType(proto.options),
        )


@attrs.define
class DataSourceSpecArgsSupplement:
    """A data class used for supplementing args protos during DataSourceSpec construction.

    This Python data class can be used to pass non-serializable types (e.g. Python functions) or data that is not
    included in args protos (e.g. schemas) into the DataSourceSpec constructor.
    """

    batch_schema: Optional[spark_schema_pb2.SparkSchema] = attrs.field(default=None)
    stream_schema: Optional[spark_schema_pb2.SparkSchema] = attrs.field(default=None)

    batch_post_processor: Optional[Callable] = attrs.field(default=None)
    stream_post_processor: Optional[Callable] = attrs.field(default=None)

    batch_data_source_function: Optional[Callable] = attrs.field(default=None)
    stream_data_source_function: Optional[Callable] = attrs.field(default=None)


@utils.frozen_strict
class BatchSourceSpec:
    """Base class for batch source specs, e.g. a HiveSourceSpec or SnowflakeSourceSpec."""

    timestamp_field: Optional[str]
    timestamp_format: Optional[str]
    post_processor: Optional[Callable] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    spark_schema: spark_schema_pb2.SparkSchema
    data_delay: Optional[pendulum.Duration]
    secrets: Optional[Mapping[str, secret_pb2.SecretReference]]


@typechecked
def create_batch_source_from_data_proto(
    proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
) -> Optional[
    Union[
        "HiveSourceSpec",
        "SparkBatchSourceSpec",
        "PandasBatchSourceSpec",
        "FileSourceSpec",
        "RedshiftSourceSpec",
        "SnowflakeSourceSpec",
        "UnitySourceSpec",
        "PushTableSourceSpec",
        "BigquerySourceSpec",
    ]
]:
    if proto.HasField("hive_table"):
        return HiveSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("unity_table"):
        return UnitySourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("spark_data_source_function"):
        return SparkBatchSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("pandas_data_source_function"):
        return PandasBatchSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("redshift_db"):
        return RedshiftSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("snowflake"):
        return SnowflakeSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("file"):
        return FileSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("push_source_table"):
        return PushTableSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("bigquery"):
        return BigquerySourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    else:
        return None


@typechecked
def create_batch_source_from_args_proto(
    proto: virtual_data_source__args_pb2.VirtualDataSourceArgs, supplement: DataSourceSpecArgsSupplement
) -> Optional[
    Union[
        "HiveSourceSpec",
        "SparkBatchSourceSpec",
        "PandasBatchSourceSpec",
        "FileSourceSpec",
        "RedshiftSourceSpec",
        "SnowflakeSourceSpec",
        "UnitySourceSpec",
        "PushTableSourceSpec",
        "BigquerySourceSpec",
    ]
]:
    if proto.HasField("hive_ds_config"):
        return HiveSourceSpec.from_args_proto(proto.hive_ds_config, supplement)
    elif proto.HasField("unity_ds_config"):
        return UnitySourceSpec.from_args_proto(proto.unity_ds_config, supplement)
    elif proto.HasField("spark_batch_config"):
        return SparkBatchSourceSpec.from_args_proto(proto.spark_batch_config, supplement)
    elif proto.HasField("pandas_batch_config"):
        return PandasBatchSourceSpec.from_args_proto(proto.pandas_batch_config, supplement)
    elif proto.HasField("redshift_ds_config"):
        return RedshiftSourceSpec.from_args_proto(proto.redshift_ds_config, supplement)
    elif proto.HasField("snowflake_ds_config"):
        return SnowflakeSourceSpec.from_args_proto(proto.snowflake_ds_config, supplement)
    elif proto.HasField("file_ds_config"):
        return FileSourceSpec.from_args_proto(proto.file_ds_config, supplement)
    elif proto.HasField("bigquery_ds_config"):
        return BigquerySourceSpec.from_args_proto(proto.bigquery_ds_config, supplement)
    else:
        return None


def _datepart_to_minimum_seconds(datepart: str) -> int:
    if datepart == "year":
        return 365 * 24 * 60 * 60
    elif datepart == "month":
        return 28 * 24 * 60 * 60
    elif datepart == "day":
        return 24 * 60 * 60
    elif datepart == "hour":
        return 60 * 60
    elif datepart == "date":
        return 24 * 60 * 60
    else:
        msg = f"Unexpected datepart string: {datepart}"
        raise ValueError(msg)


def _datepart_to_default_format_string(datepart: str) -> str:
    if datepart == "year":
        return "%Y"
    elif datepart == "month":
        return "%m"
    elif datepart == "day":
        return "%d"
    elif datepart == "hour":
        return "%H"
    elif datepart == "date":
        return "%Y-%m-%d"
    else:
        msg = f"Unexpected datepart string: {datepart}"
        raise ValueError(msg)


@utils.frozen_strict
class DatetimePartitionColumnSpec:
    column_name: str
    format_string: str
    minimum_seconds: int

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.DatetimePartitionColumn
    ) -> "DatetimePartitionColumnSpec":
        return cls(
            column_name=utils.get_field_or_none(proto, "column_name"),
            format_string=utils.get_field_or_none(proto, "format_string"),
            minimum_seconds=utils.get_field_or_none(proto, "minimum_seconds"),
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: data_source__args_pb2.DatetimePartitionColumnArgs) -> "DatetimePartitionColumnSpec":
        # This constructor mirrors backend Kotlin logic: https://github.com/tecton-ai/tecton/blob/6107d76680fa97e6155c6da9ee7d2da47bc3d6a7/java/com/tecton/datamodel/datasource/Utils.kt#L41.
        # In the short-term, divergence between Python and Kotlin will be prevented using integration test coverage. Longer-term, these values should probably be derived client-side only and included in the data source args.
        minimum_seconds = _datepart_to_minimum_seconds(proto.datepart)

        if proto.format_string:
            # NOTE: we do not use `zero_padded` if the user specified a format_string. However,
            # we do not directly validate that this is unset. We may want to in the future, but
            # not making the change now in case it causes a repo failure.
            format_string = proto.format_string
        else:
            if proto.datepart == "date" and not proto.zero_padded:
                msg = "Non-zero padded date strings are not supported. Please set `zero_padded = True` in your `DatetimePartitionColumn`."
                raise ValueError(msg)

            default_format_string = _datepart_to_default_format_string(proto.datepart)
            if proto.zero_padded:
                # The default string format should be zero padded.
                format_string = default_format_string
            else:
                format_string = default_format_string.replace("%", "%-")

        return DatetimePartitionColumnSpec(
            column_name=proto.column_name, format_string=format_string, minimum_seconds=minimum_seconds
        )


@utils.frozen_strict
class HiveSourceSpec(BatchSourceSpec):
    database: str
    table: str
    datetime_partition_columns: Tuple[DatetimePartitionColumnSpec, ...]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "HiveSourceSpec":
        # Needed for canary backwards compatability. Can be deleted on the following release.
        if proto.date_partition_column and not proto.datetime_partition_columns:
            proto.datetime_partition_columns.append(
                batch_data_source__data_pb2.DatetimePartitionColumn(
                    column_name=proto.date_partition_column,
                    format_string="%Y-%m-%d",
                    minimum_seconds=24 * 60 * 60,
                )
            )

        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            table=utils.get_field_or_none(proto.hive_table, "table"),
            database=utils.get_field_or_none(proto.hive_table, "database"),
            datetime_partition_columns=tuple(
                DatetimePartitionColumnSpec.from_data_proto(column) for column in proto.datetime_partition_columns
            ),
            secrets=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.HiveDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "HiveSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=utils.get_field_or_none(proto, "timestamp_format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            table=utils.get_field_or_none(proto, "table"),
            database=utils.get_field_or_none(proto, "database"),
            datetime_partition_columns=tuple(
                DatetimePartitionColumnSpec.from_args_proto(column) for column in proto.datetime_partition_columns
            ),
            secrets=None,
        )


@utils.frozen_strict
class PushTableSourceSpec(BatchSourceSpec):
    ingested_data_location: str

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, _: bool = False
    ) -> "PushTableSourceSpec":
        post_processor = None
        return cls(
            utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor,
            utils.get_field_or_none(proto, "spark_schema"),
            utils.get_duration_field_or_none(proto, "data_delay"),
            ingested_data_location=utils.get_field_or_none(proto.push_source_table, "ingested_data_location"),
            secrets=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.PushSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "PushTableSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=utils.get_field_or_none(proto, "timestamp_format"),
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            # TODO(Achal): Push sources don't really support this because you can't read from a
            # push source that we haven't written to and logged offline.
            ingested_data_location="",
            secrets=None,
        )


@utils.frozen_strict
class UnitySourceSpec(BatchSourceSpec):
    catalog: str
    schema: str
    table: str
    datetime_partition_columns: Tuple[DatetimePartitionColumnSpec, ...]
    access_mode: data_source__args_pb2.UnityCatalogAccessMode

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "UnitySourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)
        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            table=utils.get_field_or_none(proto.unity_table, "table"),
            schema=utils.get_field_or_none(proto.unity_table, "schema"),
            catalog=utils.get_field_or_none(proto.unity_table, "catalog"),
            datetime_partition_columns=tuple(
                DatetimePartitionColumnSpec.from_data_proto(column) for column in proto.datetime_partition_columns
            ),
            access_mode=utils.get_field_or_none(proto.unity_table, "access_mode"),
            secrets=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.UnityDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "UnitySourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor
        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=utils.get_field_or_none(proto, "timestamp_format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            table=utils.get_field_or_none(proto, "table"),
            schema=utils.get_field_or_none(proto, "schema"),
            catalog=utils.get_field_or_none(proto, "catalog"),
            datetime_partition_columns=tuple(
                DatetimePartitionColumnSpec.from_args_proto(column) for column in proto.datetime_partition_columns
            ),
            access_mode=utils.get_field_or_none(proto, "access_mode"),
            secrets=None,
        )


@utils.frozen_strict
class FileSourceSpec(BatchSourceSpec):
    uri: Optional[str]
    file_format: batch_data_source__data_pb2.FileDataSourceFormat.ValueType
    convert_to_glue_format: bool
    schema_uri: Optional[str]
    schema_override: Optional[spark_schema_pb2.SparkSchema]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "FileSourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            uri=utils.get_field_or_none(proto.file, "uri"),
            file_format=utils.get_field_or_none(proto.file, "format"),
            convert_to_glue_format=proto.file.convert_to_glue_format,
            schema_uri=utils.get_field_or_none(proto.file, "schema_uri"),
            schema_override=utils.get_field_or_none(proto.file, "schema_override"),
            secrets=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.FileDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "FileSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=utils.get_field_or_none(proto, "timestamp_format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema if supplement.batch_schema else spark_schema_pb2.SparkSchema(),
            uri=utils.get_field_or_none(proto, "uri"),
            file_format=FileSourceSpec._convert_file_format_string_to_enum(proto.file_format),
            convert_to_glue_format=proto.convert_to_glue_format,
            schema_uri=utils.get_field_or_none(proto, "schema_uri"),
            schema_override=utils.get_field_or_none(proto, "schema_override"),
            secrets=None,
        )

    @staticmethod
    @typechecked
    def _convert_file_format_string_to_enum(
        file_format: str,
    ) -> batch_data_source__data_pb2.FileDataSourceFormat.ValueType:
        return batch_data_source__data_pb2.FileDataSourceFormat.Value(f"FILE_DATA_SOURCE_FORMAT_{file_format.upper()}")


@utils.frozen_strict
class SparkBatchSourceSpec(BatchSourceSpec):
    function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    supports_time_filtering: bool

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "SparkBatchSourceSpec":
        function = None
        if proto.spark_data_source_function.HasField("function"):
            if deserialize_funcs_to_main:
                function = function_deserialization.from_proto_to_main(proto.spark_data_source_function.function)
            else:
                function = function_deserialization.from_proto(proto.spark_data_source_function.function)

        return cls(
            timestamp_field=None,
            timestamp_format=None,
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            supports_time_filtering=proto.spark_data_source_function.supports_time_filtering,
            function=function,
            secrets=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.SparkBatchConfigArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "SparkBatchSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("data_source_function"):
            data_source_function = function_deserialization.from_proto(proto.data_source_function)
        else:
            data_source_function = supplement.batch_data_source_function

        return cls(
            timestamp_field=None,
            timestamp_format=None,
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=supplement.batch_schema,
            supports_time_filtering=proto.supports_time_filtering,
            function=data_source_function,
            secrets=None,
        )


@utils.frozen_strict
class PandasBatchSourceSpec(BatchSourceSpec):
    function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    supports_time_filtering: bool

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "PandasBatchSourceSpec":
        function = None
        if proto.pandas_data_source_function.HasField("function"):
            if deserialize_funcs_to_main:
                function = function_deserialization.from_proto_to_main(proto.pandas_data_source_function.function)
            else:
                function = function_deserialization.from_proto(proto.pandas_data_source_function.function)

        return cls(
            timestamp_field=None,
            timestamp_format=None,
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            supports_time_filtering=proto.pandas_data_source_function.supports_time_filtering,
            function=function,
            secrets=proto.secrets,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.PandasBatchConfigArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "PandasBatchSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("data_source_function"):
            data_source_function = function_deserialization.from_proto(proto.data_source_function)
        else:
            data_source_function = supplement.batch_data_source_function

        return cls(
            timestamp_field=None,
            timestamp_format=None,
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=supplement.batch_schema,
            supports_time_filtering=proto.supports_time_filtering,
            function=data_source_function,
            secrets=proto.secrets,
        )


@utils.frozen_strict
class RedshiftSourceSpec(BatchSourceSpec):
    endpoint: str
    table: Optional[str]
    query: Optional[str]
    temp_s3: Optional[str]

    def __attrs_post_init__(self):
        if (self.table and self.query) or (not self.table and not self.query):
            raise errors.REDSHIFT_DS_EITHER_TABLE_OR_QUERY

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "RedshiftSourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            endpoint=utils.get_field_or_none(proto.redshift_db, "endpoint"),
            table=utils.get_field_or_none(proto.redshift_db, "table"),
            query=utils.get_field_or_none(proto.redshift_db, "query"),
            temp_s3=utils.get_field_or_none(proto.redshift_db, "temp_s3"),
            secrets=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.RedshiftDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "RedshiftSourceSpec":
        # The temp_s3 directory should be set, and this should have been verified during validation. The one exception
        # is for unit tests (i.e. `test_run()`). In unit tests, validation is skipped and the redshift directory should
        # not need to be set, so fallback to an empty string.
        temp_s3 = conf.get_or_none("SPARK_REDSHIFT_TEMP_DIR") or ""

        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=None,
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            endpoint=utils.get_field_or_none(proto, "endpoint"),
            table=utils.get_field_or_none(proto, "table"),
            query=utils.get_field_or_none(proto, "query"),
            temp_s3=temp_s3,
            secrets=None,
        )


@utils.frozen_strict
class SnowflakeSourceSpec(BatchSourceSpec):
    database: Optional[str]
    schema: Optional[str]
    warehouse: Optional[str]
    url: Optional[str]
    role: Optional[str]
    table: Optional[str]
    query: Optional[str]
    user: Optional[secret_pb2.SecretReference]
    password: Optional[secret_pb2.SecretReference]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "SnowflakeSourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)
        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            url=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "url"),
            database=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "database"),
            schema=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "schema"),
            warehouse=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "warehouse"),
            role=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "role"),
            table=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "table"),
            query=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "query"),
            secrets=None,
            user=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "user"),
            password=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "password"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.SnowflakeDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "SnowflakeSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=None,
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            url=utils.get_field_or_none(proto, "url"),
            database=utils.get_field_or_none(proto, "database"),
            schema=utils.get_field_or_none(proto, "schema"),
            warehouse=utils.get_field_or_none(proto, "warehouse"),
            role=utils.get_field_or_none(proto, "role"),
            table=utils.get_field_or_none(proto, "table"),
            query=utils.get_field_or_none(proto, "query"),
            secrets=None,
            user=utils.get_field_or_none(proto, "user"),
            password=utils.get_field_or_none(proto, "password"),
        )


@utils.frozen_strict
class BigquerySourceSpec(BatchSourceSpec):
    project_id: Optional[str]
    dataset: Optional[str]
    table: Optional[str]
    query: Optional[str]
    location: Optional[str]
    credentials: Optional[secret_pb2.SecretReference]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "BigquerySourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            project_id=utils.get_field_or_none(proto.bigquery, "project_id"),
            dataset=utils.get_field_or_none(proto.bigquery, "dataset"),
            location=utils.get_field_or_none(proto.bigquery, "location"),
            table=utils.get_field_or_none(proto.bigquery, "table"),
            query=utils.get_field_or_none(proto.bigquery, "query"),
            secrets=None,
            credentials=utils.get_field_or_none(proto.bigquery, "credentials"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.BigqueryDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "BigquerySourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=None,
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            project_id=utils.get_field_or_none(proto, "project_id"),
            dataset=utils.get_field_or_none(proto, "dataset"),
            location=utils.get_field_or_none(proto, "location"),
            table=utils.get_field_or_none(proto, "table"),
            query=utils.get_field_or_none(proto, "query"),
            secrets=None,
            credentials=utils.get_field_or_none(proto, "credentials"),
        )


@utils.frozen_strict
class StreamOptionSpec:
    key: str
    value: str


@utils.frozen_strict
class StreamSourceSpec:
    """Base class for stream source specs, e.g. a KinesisSourceSpec or KafkaSourceSpec."""

    deduplication_column_names: Tuple[str, ...]
    options: Tuple[StreamOptionSpec, ...]
    watermark_delay_threshold: pendulum.Duration
    spark_schema: spark_schema_pb2.SparkSchema


@typechecked
def create_stream_source_from_data_proto(
    proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
) -> Optional[Union["KinesisSourceSpec", "KafkaSourceSpec", "SparkStreamSourceSpec"]]:
    if proto.HasField("kinesis_data_source"):
        return KinesisSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("kafka_data_source"):
        return KafkaSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("spark_data_source_function"):
        return SparkStreamSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    else:
        return None


@typechecked
def create_stream_source_from_args_proto(
    proto: virtual_data_source__args_pb2.VirtualDataSourceArgs, supplement: DataSourceSpecArgsSupplement
) -> Optional[Union["KinesisSourceSpec", "KafkaSourceSpec", "SparkStreamSourceSpec"]]:
    if proto.HasField("kinesis_ds_config"):
        return KinesisSourceSpec.from_args_proto(proto.kinesis_ds_config, supplement)
    elif proto.HasField("kafka_ds_config"):
        return KafkaSourceSpec.from_args_proto(proto.kafka_ds_config, supplement)
    elif proto.HasField("spark_stream_config"):
        return SparkStreamSourceSpec.from_args_proto(proto.spark_stream_config, supplement)
    else:
        return None


@utils.frozen_strict
class KinesisSourceSpec(StreamSourceSpec):
    post_processor: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    stream_name: str
    region: str
    initial_stream_position: data_source_config_pb2.InitialStreamPosition.ValueType

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
    ) -> "KinesisSourceSpec":
        post_processor = None
        if proto.HasField("raw_stream_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_stream_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_stream_translator)

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.deduplication_column_names),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.stream_config.watermark_delay_threshold),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            stream_name=utils.get_field_or_none(proto.kinesis_data_source, "stream_name"),
            region=utils.get_field_or_none(proto.kinesis_data_source, "region"),
            initial_stream_position=proto.stream_config.initial_stream_position,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.KinesisDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "KinesisSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.stream_post_processor

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.common_args.deduplication_columns),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.common_args.watermark_delay_threshold),
            spark_schema=supplement.stream_schema,
            stream_name=utils.get_field_or_none(proto, "stream_name"),
            region=utils.get_field_or_none(proto, "region"),
            initial_stream_position=proto.initial_stream_position,
        )


@utils.frozen_strict
class KafkaSourceSpec(StreamSourceSpec):
    post_processor: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    bootstrap_servers: str
    topics: str
    ssl_keystore_location: Optional[str]
    ssl_keystore_password_secret_id: Optional[str]
    ssl_truststore_location: Optional[str]
    ssl_truststore_password_secret_id: Optional[str]
    security_protocol: Optional[str]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
    ) -> "KafkaSourceSpec":
        post_processor = None
        if proto.HasField("raw_stream_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_stream_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_stream_translator)

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.deduplication_column_names),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.stream_config.watermark_delay_threshold),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            bootstrap_servers=utils.get_field_or_none(proto.kafka_data_source, "bootstrap_servers"),
            topics=utils.get_field_or_none(proto.kafka_data_source, "topics"),
            ssl_keystore_location=utils.get_field_or_none(proto.kafka_data_source, "ssl_keystore_location"),
            ssl_keystore_password_secret_id=utils.get_field_or_none(
                proto.kafka_data_source, "ssl_keystore_password_secret_id"
            ),
            ssl_truststore_location=utils.get_field_or_none(proto.kafka_data_source, "ssl_truststore_location"),
            ssl_truststore_password_secret_id=utils.get_field_or_none(
                proto.kafka_data_source, "ssl_truststore_password_secret_id"
            ),
            security_protocol=utils.get_field_or_none(proto.kafka_data_source, "security_protocol"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.KafkaDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "KafkaSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.stream_post_processor

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.common_args.deduplication_columns),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.common_args.watermark_delay_threshold),
            spark_schema=supplement.stream_schema,
            bootstrap_servers=utils.get_field_or_none(proto, "kafka_bootstrap_servers"),
            topics=utils.get_field_or_none(proto, "topics"),
            ssl_keystore_location=utils.get_field_or_none(proto, "ssl_keystore_location"),
            ssl_keystore_password_secret_id=utils.get_field_or_none(proto, "ssl_keystore_password_secret_id"),
            ssl_truststore_location=utils.get_field_or_none(proto, "ssl_truststore_location"),
            ssl_truststore_password_secret_id=utils.get_field_or_none(proto, "ssl_truststore_password_secret_id"),
            security_protocol=utils.get_field_or_none(proto, "security_protocol"),
        )


@utils.frozen_strict
class SparkStreamSourceSpec(StreamSourceSpec):
    function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
    ) -> "SparkStreamSourceSpec":
        function = None
        if proto.spark_data_source_function.HasField("function"):
            if deserialize_funcs_to_main:
                function = function_deserialization.from_proto_to_main(proto.spark_data_source_function.function)
            else:
                function = function_deserialization.from_proto(proto.spark_data_source_function.function)

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.deduplication_column_names),
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.stream_config.watermark_delay_threshold),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            function=function,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.SparkStreamConfigArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "SparkStreamSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("data_source_function"):
            data_source_function = function_deserialization.from_proto(proto.data_source_function)
        else:
            data_source_function = supplement.stream_data_source_function

        return cls(
            deduplication_column_names=(),
            options=(),
            watermark_delay_threshold=pendulum.Duration(),
            spark_schema=supplement.stream_schema,
            function=data_source_function,
        )


# Resolve forward type declarations.
attrs.resolve_types(DataSourceSpec, locals(), globals())
attrs.resolve_types(StreamSourceSpec, locals(), globals())



import glob
import os
import sys
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import click
import pytest

from tecton.cli import printer
from tecton.cli.command import TectonCommand
from tecton.cli.repo_utils import get_tecton_objects_skip_validation
from tecton_core import conf
from tecton_core import repo_file_handler


ALLOWED_COMPUTE_MODES = ["rift", "spark"]


def get_test_paths(repo_root) -> List[str]:
    # Be _very_ careful updating this:
    #    `glob.glob` does bash-style globbing (ignores hidden files)
    #    `pathlib.Path.glob` does _not_ do bash-style glob (it shows hidden)
    #
    # Ignoring hidden files is a very important expectation for our usage of
    # pytest. Otherwise, we may test files that user does not intend us to
    # (like in their .git or .tox directories).
    #
    # NOTE: This won't filter out hidden files for Windows. Potentially:
    #    `bool(os.stat(filepath).st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN)`
    # would filter hidden files for Windows, but this would need some testing.
    candidate_test_files = glob.iglob(f"{repo_root}/**/tests/**/*.py", recursive=True)

    VIRTUAL_ENV = os.getenv("VIRTUAL_ENV")
    if VIRTUAL_ENV:
        return list(filter(lambda f: not f.startswith(VIRTUAL_ENV), candidate_test_files))

    return list(candidate_test_files)


def run_tests(repo_config_path: Optional[Path], pytest_extra_args: Tuple[str, ...] = ()):
    repo_root = repo_file_handler._maybe_get_repo_root()
    if repo_root is None:
        printer.safe_print("Tecton tests must be run from a feature repo initialized using 'tecton init'!")
        sys.exit(1)

    get_tecton_objects_skip_validation(repo_config_path)

    tests = get_test_paths(repo_root)
    if len(tests) == 0:
        printer.safe_print("⚠️  Running Tests: No tests found.")
        return

    os.chdir(repo_root)
    args = ["--disable-pytest-warnings", "-s", *tests]

    if pytest_extra_args:
        args.extend(pytest_extra_args)

    printer.safe_print("🏃 Running Tests")
    exitcode = pytest.main(args)

    if exitcode == 5:
        # https://docs.pytest.org/en/stable/usage.html#possible-exit-codes
        printer.safe_print("⚠️  Running Tests: No tests found.")
        return None
    elif exitcode != 0:
        printer.safe_print("⛔ Running Tests: Tests failed :(")
        sys.exit(1)
    else:
        printer.safe_print("✅ Running Tests: Tests passed!")


@click.command(uses_workspace=True, requires_auth=False, cls=TectonCommand)
@click.option(
    "--enable-python-serialization/--disable-python-serialization",
    show_default=True,
    is_flag=True,
    default=True,
    help="""
    If disabled, Tecton will not serialize python code during unit tests. This can be useful in some test environments
    or when running code coverage tools, however the tests may be less realistic since serialization issues will not be
    covered. This option is not supported when running tests during `tecton apply`. If using pytest directly, set
    TECTON_FORCE_FUNCTION_SERIALIZATION=false in your environment to achieve the same behavior.
    """,
)
@click.option(
    "--config",
    help="Path to the repo config yaml file. Defaults to the repo.yaml file at the Tecton repo root.",
    default=None,
    type=click.Path(exists=True, dir_okay=False, path_type=Path, readable=True),
)
@click.option(
    "--default-compute-mode",
    help="What compute mode to use to test features if not specified. Should one of spark or rift.",
    default="spark",
    type=click.Choice(choices=ALLOWED_COMPUTE_MODES, case_sensitive=False),
)
@click.argument("pytest_extra_args", nargs=-1)
def test(
    enable_python_serialization,
    config: Optional[Path],
    default_compute_mode: str,
    pytest_extra_args: Tuple[str, ...],
):
    """Run Tecton tests.
    USAGE:
    `tecton test`: run all tests (using PyTest) in a file that matches glob("TECTON_REPO_ROOT/**/tests/**/*.py")
    `tecton test -- -k "test_name"`: same as above, but passes the `-k "test_name"` args to the PyTest command.
    """
    if conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION"):
        msg = "Do not set TECTON_FORCE_FUNCTION_SERIALIZATION when using `tecton test`. Use --enable-python-serialization/--disable-python-serialization instead."
        raise RuntimeError(msg)

    if enable_python_serialization:
        conf.set("TECTON_FORCE_FUNCTION_SERIALIZATION", "true")
    else:
        conf.set("TECTON_FORCE_FUNCTION_SERIALIZATION", "false")

    default_compute_mode = default_compute_mode.lower()
    offline_retrieval_compute_mode = (
        conf._get_runtime_only("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE") or default_compute_mode
    )
    batch_compute_mode = conf._get_runtime_only("TECTON_BATCH_COMPUTE_MODE") or default_compute_mode
    # These values are set explicitly in unit tests in order to avoid needing to be logged into a cluster
    # to run unit tests. Without this, the compute_mode is derived from the user's cluster, which requires log-in.
    with conf._temporary_set(
        "TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", offline_retrieval_compute_mode
    ) as a, conf._temporary_set("TECTON_BATCH_COMPUTE_MODE", batch_compute_mode) as b:
        # NOTE: if a user wanted to do the equivalent of a `pytest -k "test_name"`
        # they could do `tecton test -- -k "test_name"`.
        run_tests(
            repo_config_path=config,
            pytest_extra_args=pytest_extra_args,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/modelartifactservice/model_artifact_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nFtecton_proto/modelartifactservice/model_artifact_service__client.proto\x12!tecton_proto.modelartifactservice\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\'tecton_proto/auth/service__client.proto\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\"\xfa\x06\n\x11ModelArtifactInfo\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12@\n\x04type\x18\x03 \x01(\x0e\x32,.tecton_proto.modelartifactservice.ModelTypeR\x04type\x12 \n\x0b\x64\x65scription\x18\x04 \x01(\tR\x0b\x64\x65scription\x12\x65\n\x0b\x66ile_hashes\x18\x05 \x03(\x0b\x32\x44.tecton_proto.modelartifactservice.ModelArtifactInfo.FileHashesEntryR\nfileHashes\x12&\n\x0fmodel_file_path\x18\r \x01(\tR\rmodelFilePath\x12>\n\x0cinput_schema\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0binputSchema\x12@\n\routput_schema\x18\x07 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0coutputSchema\x12R\n\x04tags\x18\t \x03(\x0b\x32>.tecton_proto.modelartifactservice.ModelArtifactInfo.TagsEntryR\x04tags\x12!\n\x0cstorage_path\x18\x08 \x01(\tR\x0bstoragePath\x12N\n\x06status\x18\n \x01(\x0e\x32\x36.tecton_proto.modelartifactservice.ModelArtifactStatusR\x06status\x12\x39\n\ncreated_at\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x39\n\nupdated_at\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAt\x1a=\n\x0f\x46ileHashesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"X\n\x19ListModelArtifactsRequest\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\"j\n\x1aListModelArtifactsResponse\x12L\n\x06models\x18\x01 \x03(\x0b\x32\x34.tecton_proto.modelartifactservice.ModelArtifactInfoR\x06models\"[\n\x1c\x44\x65scribeModelArtifactRequest\x12\'\n\x02id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"\x85\x01\n\x1d\x44\x65scribeModelArtifactResponse\x12\x64\n\x13model_artifact_info\x18\x01 \x01(\x0b\x32\x34.tecton_proto.modelartifactservice.ModelArtifactInfoR\x11modelArtifactInfo\"X\n\x19\x46\x65tchModelArtifactRequest\x12\'\n\x02id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"?\n\x1a\x46\x65tchModelArtifactResponse\x12!\n\x0c\x64ownload_url\x18\x01 \x01(\tR\x0b\x64ownloadUrl\"\x83\x05\n\x1a\x43reateModelArtifactRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12@\n\x04type\x18\x02 \x01(\x0e\x32,.tecton_proto.modelartifactservice.ModelTypeR\x04type\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12n\n\x0b\x66ile_hashes\x18\x04 \x03(\x0b\x32M.tecton_proto.modelartifactservice.CreateModelArtifactRequest.FileHashesEntryR\nfileHashes\x12&\n\x0fmodel_file_path\x18\x08 \x01(\tR\rmodelFilePath\x12>\n\x0cinput_schema\x18\x05 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0binputSchema\x12@\n\routput_schema\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x0coutputSchema\x12[\n\x04tags\x18\x07 \x03(\x0b\x32G.tecton_proto.modelartifactservice.CreateModelArtifactRequest.TagsEntryR\x04tags\x1a=\n\x0f\x46ileHashesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\x83\x01\n\x1b\x43reateModelArtifactResponse\x12\x64\n\x13model_artifact_info\x18\x01 \x01(\x0b\x32\x34.tecton_proto.modelartifactservice.ModelArtifactInfoR\x11modelArtifactInfo\"g\n GetModelArtifactUploadUrlRequest\x12\x43\n\x11model_artifact_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fmodelArtifactId\"\x85\x01\n!GetModelArtifactUploadUrlResponse\x12\x43\n\x11model_artifact_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fmodelArtifactId\x12\x1b\n\tupload_id\x18\x02 \x01(\tR\x08uploadId\"\xb0\x01\n\x1eUploadModelArtifactPartRequest\x12\x43\n\x11model_artifact_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fmodelArtifactId\x12(\n\x10parent_upload_id\x18\x02 \x01(\tR\x0eparentUploadId\x12\x1f\n\x0bpart_number\x18\x03 \x01(\x05R\npartNumber\"@\n\x1fUploadModelArtifactPartResponse\x12\x1d\n\nupload_url\x18\x01 \x01(\tR\tuploadUrl\"\xb9\x02\n\"CompleteModelArtifactUploadRequest\x12\x43\n\x11model_artifact_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x0fmodelArtifactId\x12\x1b\n\tupload_id\x18\x02 \x01(\tR\x08uploadId\x12s\n\npart_etags\x18\x03 \x03(\x0b\x32T.tecton_proto.modelartifactservice.CompleteModelArtifactUploadRequest.PartEtagsEntryR\tpartEtags\x1a<\n\x0ePartEtagsEntry\x12\x10\n\x03key\x18\x01 \x01(\x05R\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"7\n#CompleteModelArtifactUploadResponse\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\"Y\n\x1a\x44\x65leteModelArtifactRequest\x12\'\n\x02id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\"\x1d\n\x1b\x44\x65leteModelArtifactResponse*t\n\x13ModelArtifactStatus\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x10\n\x0cPENDING_FILE\x10\x01\x12\x10\n\x0cPENDING_SCAN\x10\x02\x12\t\n\x05READY\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\x0b\n\x07\x44\x45LETED\x10\x05*O\n\tModelType\x12\x1a\n\x16MODEL_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15TECTON_TEXT_EMBEDDING\x10\x01\x12\x0b\n\x07PYTORCH\x10\x02\x32\xc5\x0b\n\x14ModelArtifactService\x12\xc6\x01\n\x13\x43reateModelArtifact\x12=.tecton_proto.modelartifactservice.CreateModelArtifactRequest\x1a>.tecton_proto.modelartifactservice.CreateModelArtifactResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xc3\x01\n\x12ListModelArtifacts\x12<.tecton_proto.modelartifactservice.ListModelArtifactsRequest\x1a=.tecton_proto.modelartifactservice.ListModelArtifactsResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xc3\x01\n\x12\x46\x65tchModelArtifact\x12<.tecton_proto.modelartifactservice.FetchModelArtifactRequest\x1a=.tecton_proto.modelartifactservice.FetchModelArtifactResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xd8\x01\n\x19GetModelArtifactUploadUrl\x12\x43.tecton_proto.modelartifactservice.GetModelArtifactUploadUrlRequest\x1a\x44.tecton_proto.modelartifactservice.GetModelArtifactUploadUrlResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xd2\x01\n\x17UploadModelArtifactPart\x12\x41.tecton_proto.modelartifactservice.UploadModelArtifactPartRequest\x1a\x42.tecton_proto.modelartifactservice.UploadModelArtifactPartResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xde\x01\n\x1b\x43ompleteModelArtifactUpload\x12\x45.tecton_proto.modelartifactservice.CompleteModelArtifactUploadRequest\x1a\x46.tecton_proto.modelartifactservice.CompleteModelArtifactUploadResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_server\x12\xc6\x01\n\x13\x44\x65leteModelArtifact\x12=.tecton_proto.modelartifactservice.DeleteModelArtifactRequest\x1a>.tecton_proto.modelartifactservice.DeleteModelArtifactResponse\"0\xa2\xbc\xe6\xc0\x05**(tecton/organization#scale_feature_serverBv\n\x1f\x63om.tecton.modelartifactserviceB\x19ModelArtifactServiceProtoP\x01Z6github.com/tecton-ai/tecton_proto/modelartifactservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.modelartifactservice.model_artifact_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.tecton.modelartifactserviceB\031ModelArtifactServiceProtoP\001Z6github.com/tecton-ai/tecton_proto/modelartifactservice'
  _MODELARTIFACTINFO_FILEHASHESENTRY._options = None
  _MODELARTIFACTINFO_FILEHASHESENTRY._serialized_options = b'8\001'
  _MODELARTIFACTINFO_TAGSENTRY._options = None
  _MODELARTIFACTINFO_TAGSENTRY._serialized_options = b'8\001'
  _CREATEMODELARTIFACTREQUEST_FILEHASHESENTRY._options = None
  _CREATEMODELARTIFACTREQUEST_FILEHASHESENTRY._serialized_options = b'8\001'
  _CREATEMODELARTIFACTREQUEST_TAGSENTRY._options = None
  _CREATEMODELARTIFACTREQUEST_TAGSENTRY._serialized_options = b'8\001'
  _COMPLETEMODELARTIFACTUPLOADREQUEST_PARTETAGSENTRY._options = None
  _COMPLETEMODELARTIFACTUPLOADREQUEST_PARTETAGSENTRY._serialized_options = b'8\001'
  _MODELARTIFACTSERVICE.methods_by_name['CreateModelArtifact']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['CreateModelArtifact']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSERVICE.methods_by_name['ListModelArtifacts']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['ListModelArtifacts']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSERVICE.methods_by_name['FetchModelArtifact']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['FetchModelArtifact']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSERVICE.methods_by_name['GetModelArtifactUploadUrl']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['GetModelArtifactUploadUrl']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSERVICE.methods_by_name['UploadModelArtifactPart']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['UploadModelArtifactPart']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSERVICE.methods_by_name['CompleteModelArtifactUpload']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['CompleteModelArtifactUpload']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSERVICE.methods_by_name['DeleteModelArtifact']._options = None
  _MODELARTIFACTSERVICE.methods_by_name['DeleteModelArtifact']._serialized_options = b'\242\274\346\300\005**(tecton/organization#scale_feature_server'
  _MODELARTIFACTSTATUS._serialized_start=3499
  _MODELARTIFACTSTATUS._serialized_end=3615
  _MODELTYPE._serialized_start=3617
  _MODELTYPE._serialized_end=3696
  _MODELARTIFACTINFO._serialized_start=264
  _MODELARTIFACTINFO._serialized_end=1154
  _MODELARTIFACTINFO_FILEHASHESENTRY._serialized_start=1036
  _MODELARTIFACTINFO_FILEHASHESENTRY._serialized_end=1097
  _MODELARTIFACTINFO_TAGSENTRY._serialized_start=1099
  _MODELARTIFACTINFO_TAGSENTRY._serialized_end=1154
  _LISTMODELARTIFACTSREQUEST._serialized_start=1156
  _LISTMODELARTIFACTSREQUEST._serialized_end=1244
  _LISTMODELARTIFACTSRESPONSE._serialized_start=1246
  _LISTMODELARTIFACTSRESPONSE._serialized_end=1352
  _DESCRIBEMODELARTIFACTREQUEST._serialized_start=1354
  _DESCRIBEMODELARTIFACTREQUEST._serialized_end=1445
  _DESCRIBEMODELARTIFACTRESPONSE._serialized_start=1448
  _DESCRIBEMODELARTIFACTRESPONSE._serialized_end=1581
  _FETCHMODELARTIFACTREQUEST._serialized_start=1583
  _FETCHMODELARTIFACTREQUEST._serialized_end=1671
  _FETCHMODELARTIFACTRESPONSE._serialized_start=1673
  _FETCHMODELARTIFACTRESPONSE._serialized_end=1736
  _CREATEMODELARTIFACTREQUEST._serialized_start=1739
  _CREATEMODELARTIFACTREQUEST._serialized_end=2382
  _CREATEMODELARTIFACTREQUEST_FILEHASHESENTRY._serialized_start=1036
  _CREATEMODELARTIFACTREQUEST_FILEHASHESENTRY._serialized_end=1097
  _CREATEMODELARTIFACTREQUEST_TAGSENTRY._serialized_start=1099
  _CREATEMODELARTIFACTREQUEST_TAGSENTRY._serialized_end=1154
  _CREATEMODELARTIFACTRESPONSE._serialized_start=2385
  _CREATEMODELARTIFACTRESPONSE._serialized_end=2516
  _GETMODELARTIFACTUPLOADURLREQUEST._serialized_start=2518
  _GETMODELARTIFACTUPLOADURLREQUEST._serialized_end=2621
  _GETMODELARTIFACTUPLOADURLRESPONSE._serialized_start=2624
  _GETMODELARTIFACTUPLOADURLRESPONSE._serialized_end=2757
  _UPLOADMODELARTIFACTPARTREQUEST._serialized_start=2760
  _UPLOADMODELARTIFACTPARTREQUEST._serialized_end=2936
  _UPLOADMODELARTIFACTPARTRESPONSE._serialized_start=2938
  _UPLOADMODELARTIFACTPARTRESPONSE._serialized_end=3002
  _COMPLETEMODELARTIFACTUPLOADREQUEST._serialized_start=3005
  _COMPLETEMODELARTIFACTUPLOADREQUEST._serialized_end=3318
  _COMPLETEMODELARTIFACTUPLOADREQUEST_PARTETAGSENTRY._serialized_start=3258
  _COMPLETEMODELARTIFACTUPLOADREQUEST_PARTETAGSENTRY._serialized_end=3318
  _COMPLETEMODELARTIFACTUPLOADRESPONSE._serialized_start=3320
  _COMPLETEMODELARTIFACTUPLOADRESPONSE._serialized_end=3375
  _DELETEMODELARTIFACTREQUEST._serialized_start=3377
  _DELETEMODELARTIFACTREQUEST._serialized_end=3466
  _DELETEMODELARTIFACTRESPONSE._serialized_start=3468
  _DELETEMODELARTIFACTRESPONSE._serialized_end=3497
  _MODELARTIFACTSERVICE._serialized_start=3699
  _MODELARTIFACTSERVICE._serialized_end=5176
# @@protoc_insertion_point(module_scope)



from collections import defaultdict
from datetime import datetime
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import pyspark
import pyspark.sql.functions as F
from pyspark.sql import Column

from tecton_core.aggregation_utils import get_materialization_aggregation_column_prefixes
from tecton_core.aggregation_utils import get_pretty_column_prefix
from tecton_core.query import compaction_utils
from tecton_core.query_consts import aggregation_group_id
from tecton_core.query_consts import aggregation_tile_id
from tecton_core.query_consts import anchor_time
from tecton_core.schema import Schema
from tecton_proto.data.feature_view__client_pb2 import TrailingTimeWindowAggregation
from tecton_spark.aggregation_plans import AggregationPlan
from tecton_spark.aggregation_plans import get_aggregation_plan
from tecton_spark.time_utils import convert_timestamp_to_epoch
from tecton_spark.time_utils import get_timestamp_in_seconds


WINDOW_COLUMN_NAME = "window"


def _get_feature_partial_aggregations(
    aggregation_plan: AggregationPlan, feature_name: str
) -> Iterator[Tuple[str, Column]]:
    column_names = set()
    input_columns = (
        aggregation_plan.intermediate_column_names(feature_name)
        if aggregation_plan.intermediate_column_prefixes is not None
        else [feature_name]
    )

    for column_name, aggregated_column in zip(
        aggregation_plan.materialized_column_names(feature_name),
        aggregation_plan.partial_aggregation_transform(input_columns),
    ):
        if column_name in column_names:
            continue
        column_names.add(column_name)

        yield column_name, aggregated_column.alias(column_name)


def _convert_window_to_anchor_time(
    output_df: pyspark.sql.DataFrame,
    is_continuous: bool,
    time_key: str,
    version: int,
    window_start_column_name: Optional[str],
    window_end_column_name: Optional[str],
    convert_window_times_to_epoch: bool,
) -> pyspark.sql.DataFrame:
    def _add_time_column(
        df: pyspark.sql.DataFrame, input_ts_column_name: str, output_column_name: str
    ) -> pyspark.sql.DataFrame:
        col = F.col(input_ts_column_name)
        return df.withColumn(
            output_column_name, convert_timestamp_to_epoch(col, version) if convert_window_times_to_epoch else col
        )

    # For continuous aggregations this will simply be the time key.
    if is_continuous:
        return _add_time_column(output_df, time_key, anchor_time())

    # Grouping by Spark Window introduces the "window" struct with "start" and "end" columns.
    # We only need to keep the "start" column as an anchor time.
    anchor_column_name = window_start_column_name if window_start_column_name else anchor_time()
    output_df = _add_time_column(output_df, f"{WINDOW_COLUMN_NAME}.start", anchor_column_name)

    if window_end_column_name:
        output_df = _add_time_column(output_df, f"{WINDOW_COLUMN_NAME}.end", window_end_column_name)

    return output_df.drop(WINDOW_COLUMN_NAME)


def construct_partial_time_aggregation_df(
    df: pyspark.sql.DataFrame,
    group_by_columns: List[str],
    time_aggregation: TrailingTimeWindowAggregation,
    version: int,
    view_schema: Schema,
    window_start_column_name: Optional[str] = None,
    window_end_column_name: Optional[str] = None,
    convert_to_epoch: bool = True,
    aggregation_anchor_time: Optional[datetime] = None,
) -> pyspark.sql.DataFrame:
    """Constructs a dataframe that performs partial aggregations on the input dataframe.

    Also removes the default window column that is generated, and replaces it with a start time column, and optionally
    an end time column.

    If the time aggregation is continuous, no aggregations will be performed. Instead, a start time column will be added
    but will simply be the same as the timestamp column in the input dataframe.

    Args:
        df: The input dataframe.
        group_by_columns: The columns to be grouped by for the partial aggregations.
        time_aggregation: The set of partial aggregations to be performed.
        version: The feature store version.
        window_start_column_name: If specified, the name of the start time column; otherwise it will use the default value "_anchor_time".
        window_end_column_name: If specified, the end time column will be included with this name.
        convert_to_epoch: If True, the window start and end times will be converted to epoch.
        aggregation_anchor_time: If specified, the offset for the aggregation windows.
        view_schema: The schema of the dataframe the partial aggregations are being performed for.

    Returns:
        A dataframe with the partial aggregations. For example, if `user_id` is the only join key, there are two
        aggregations (min and max), `window_start_column_name` is `tile_start_time`, `window_end_column_name` is
        `tile_end_time`, and `convert_to_epoch` is `False`, then the resulting dataframe might look something like:

        +-------+------------+------------+-------------------+-------------------+
        |user_id|value_min_1d|value_max_1d|    tile_start_time|      tile_end_time|
        +-------+------------+------------+-------------------+-------------------+
        |      1|           7|           9|2022-05-15 00:00:00|2022-05-16 00:00:00|
        |      2|          10|          12|2022-05-16 00:00:00|2022-05-17 00:00:00|
        |      1|          13|          13|2022-05-16 00:00:00|2022-05-17 00:00:00|
        +-------+------------+------------+-------------------+-------------------+
    """
    output_columns = set()
    # If create_tiles is None, default to creating tiles whenever the fv is not using continuous mode
    aggregation_tile_interval = time_aggregation.aggregation_slide_period.seconds
    create_tiles = aggregation_tile_interval > 0
    if create_tiles:
        group_by_cols = [F.col(col) for col in group_by_columns]
        slide_str = f"{aggregation_tile_interval} seconds"

        anchor_time_offset_string = None
        if aggregation_anchor_time:
            # Compute the offset from the epoch such that anchor_time aligns to an interval boundary of size
            # aggregation_slide_period. i.e. `epoch + offset + (X * slide_period) = anchor_time`, where X is
            # an integer.
            anchor_time_epoch = get_timestamp_in_seconds(aggregation_anchor_time)
            slide_period_seconds = aggregation_tile_interval
            anchor_time_offset_seconds = anchor_time_epoch % slide_period_seconds
            anchor_time_offset_string = f"{anchor_time_offset_seconds} seconds"

        window_spec = F.window(time_aggregation.time_key, slide_str, slide_str, anchor_time_offset_string)
        group_by_cols = [window_spec, *group_by_cols]
        aggregations = []
        intermediate_columns_added = set()
        for feature in time_aggregation.features:
            col_datatype = view_schema.to_dict()[feature.input_feature_name]
            aggregation_plan = get_aggregation_plan(
                feature.function,
                feature.function_params,
                time_aggregation.is_continuous,
                time_aggregation.time_key,
                col_datatype,
            )

            if aggregation_plan.partial_aggregation_preprocessor is not None:
                intermediate_columns = aggregation_plan.partial_aggregation_preprocessor(feature.input_feature_name)
                intermediate_column_names = aggregation_plan.intermediate_column_names(feature.input_feature_name)
                for column, column_name in zip(intermediate_columns, intermediate_column_names):
                    if column_name not in intermediate_columns_added:
                        intermediate_columns_added.add(column_name)
                        df = df.withColumn(column_name, column)

            for name, aggregation in _get_feature_partial_aggregations(aggregation_plan, feature.input_feature_name):
                if name in output_columns:
                    continue
                output_columns.add(name)
                aggregations.append(aggregation)
        output_df = df.groupBy(*group_by_cols).agg(*aggregations)
    else:
        columns_to_drop = set()
        intermediate_columns_added = set()
        for feature in time_aggregation.features:
            col_datatype = view_schema.to_dict()[feature.input_feature_name]

            aggregation_plan = get_aggregation_plan(
                feature.function,
                feature.function_params,
                time_aggregation.is_continuous,
                time_aggregation.time_key,
                col_datatype,
            )

            if aggregation_plan.partial_aggregation_preprocessor is not None:
                intermediate_columns = aggregation_plan.partial_aggregation_preprocessor(feature.input_feature_name)
                intermediate_column_names = aggregation_plan.intermediate_column_names(feature.input_feature_name)
                for column, column_name in zip(intermediate_columns, intermediate_column_names):
                    if column_name not in intermediate_columns_added:
                        intermediate_columns_added.add(column_name)
                        df = df.withColumn(column_name, column)

            input_columns = (
                aggregation_plan.intermediate_column_names(feature.input_feature_name)
                if aggregation_plan.intermediate_column_prefixes is not None
                else [feature.input_feature_name]
            )
            continuous_columns = aggregation_plan.continuous_partial_aggregation_transform(input_columns)

            continuous_column_prefixes = get_materialization_aggregation_column_prefixes(
                feature.function, function_params=feature.function_params, is_continuous=time_aggregation.is_continuous
            )
            continuous_column_names = [
                f"{column_prefix}_{feature.input_feature_name}" for column_prefix in continuous_column_prefixes
            ]

            for column_name, column in zip(continuous_column_names, continuous_columns):
                if column_name in output_columns:
                    continue
                output_columns.add(column_name)
                df = df.withColumn(column_name, column)
            columns_to_drop.add(feature.input_feature_name)
            columns_to_drop.update(input_columns)
        # Drop the original feature columns.
        for column in columns_to_drop:
            df = df.drop(column)
        output_df = df

    output_df = _convert_window_to_anchor_time(
        output_df,
        not create_tiles,
        time_aggregation.time_key,
        version,
        window_start_column_name,
        window_end_column_name,
        convert_to_epoch,
    )
    # TOOD: TEC-12299 drop the `timestamp` column in m13n schema for continuous WAFV
    return output_df


def partial_aggregate_column_renames(
    slide_interval_string: str, trailing_time_window_aggregation: TrailingTimeWindowAggregation
) -> Dict[str, str]:
    """Rename partial aggregate columns to human readable format."""
    # Create a map from intermediate rollup column name to preferred column names.
    renaming_map = {}
    for feature in trailing_time_window_aggregation.features:
        aggregation_plan = get_aggregation_plan(
            feature.function,
            feature.function_params,
            trailing_time_window_aggregation.is_continuous,
            trailing_time_window_aggregation.time_key,
        )

        for materialized_column_prefix, materialized_column_name in zip(
            aggregation_plan.materialized_column_prefixes,
            aggregation_plan.materialized_column_names(feature.input_feature_name),
        ):
            new_prefix = get_pretty_column_prefix(materialized_column_prefix)
            renaming_map[materialized_column_name] = (
                f"{feature.input_feature_name}_{new_prefix}_{slide_interval_string}"
            )
    return renaming_map


def construct_online_partial_agg_v2_df(
    df: pyspark.sql.DataFrame,
    group_by_columns: List[str],
    aggregation_groups: Tuple[compaction_utils.AggregationGroup, ...],
    time_key: str,
    view_schema: Schema,
    is_continuous: bool = False,
) -> pyspark.sql.DataFrame:
    """Constructs a dataframe that performs partial aggregations for online materialization on the input dataframe.

    This is built to follow the pattern used for the batch-compacted online store format. It does the aggregations
    specified by the `aggregation_groups`, grouped by the `group_by_columns`. Rows in the input dataframe are already mapped
    to the appropriate aggregation_group via the `AGGREGATION_GROUP_ID` column.
    """

    group_by_cols = [F.col(col) for col in group_by_columns]

    aggregations = []

    intermediate_columns_added = set()

    for agg_group in aggregation_groups:
        sub_aggs = []
        sub_outputs = set()

        for feature in agg_group.aggregate_features:
            col_datatype = view_schema.to_dict()[feature.input_feature_name]
            aggregation_plan = get_aggregation_plan(
                feature.function, feature.function_params, is_continuous, time_key, col_datatype
            )

            # partial_aggregation_preprocessor is used by approximate count distinct
            if aggregation_plan.partial_aggregation_preprocessor is not None:
                intermediate_columns = aggregation_plan.partial_aggregation_preprocessor(feature.input_feature_name)
                intermediate_column_names = aggregation_plan.intermediate_column_names(feature.input_feature_name)
                for column, column_name in zip(intermediate_columns, intermediate_column_names):
                    if column_name not in intermediate_columns_added:
                        intermediate_columns_added.add(column_name)
                        df = df.withColumn(column_name, column)

            for name, aggregation in _get_feature_partial_aggregations(aggregation_plan, feature.input_feature_name):
                # Skip duplicated partial aggregates. This can happen if you have mean + count over the same column, since they both use a count partial aggregate.
                if name in sub_outputs:
                    continue
                sub_outputs.add(name)
                sub_aggs.append(aggregation)

        expected_columns = agg_group.schema.column_names()
        if len(sub_outputs) != len(expected_columns):
            msg = "unexpected difference between expected columns and output columns"
            raise ValueError(msg)
        if set(sub_outputs) != set(expected_columns):
            msg = "unexpected difference between expected columns and output columns"
            raise ValueError(msg)

        aggregations.append(
            F.when(
                (F.col(aggregation_group_id()) == agg_group.window_index)
                & (F.col(aggregation_tile_id()) == agg_group.tile_index),
                F.struct(*sub_aggs),
            )
            .otherwise(F.lit(None))
            .alias(agg_group.window_tile_column_name)
        )

    # Compute aggregations for each aggregation window and tile.
    output_df = df.groupBy(*group_by_cols).agg(*aggregations)
    return output_df


def construct_partial_agg_lists_for_online_df(
    df: pyspark.sql.DataFrame,
    group_by_columns: List[str],
    aggregation_groups: Tuple[compaction_utils.AggregationGroup, ...],
) -> pyspark.sql.DataFrame:
    """Takes input data containing partial aggregations and collects all partial aggregations for a given aggregation window into one list.

    This is built to follow the pattern used for the batch-compacted online store format and used for the online compaction query.
    """
    group_by_cols = [F.col(col) for col in group_by_columns]

    input_tile_columns = []
    first_agg_columns = []

    agg_window_indexes_to_agg_groups = defaultdict(list)

    for agg_group in aggregation_groups:
        agg_window_indexes_to_agg_groups[agg_group.window_index].append(agg_group)

        aggregation_column_for_tile = F.col(agg_group.window_tile_column_name)
        input_tile_columns.append(aggregation_column_for_tile)
        # Every row corresponds to only 1 agg id + tile id combination, and sets nulls for the other tile columns. Use ignorenulls=True so we only select the first (and only) non-null value for that tile.
        first_agg_columns.append(
            F.first(aggregation_column_for_tile, ignorenulls=True).alias(agg_group.window_tile_column_name)
        )

    # For each join key and aggregation window, there is only 1 non-null value in each aggregation tile column across all the rows. This is why we can use first() aggregation to get the tile agg values.
    output_df = df.select(*group_by_cols, *input_tile_columns).groupBy(*group_by_cols).agg(*first_agg_columns)

    # Collapse all tile columns into one array with the tiles sorted in ascending order. The ordering is important.
    array_agg_columns = []
    for agg_window in agg_window_indexes_to_agg_groups.keys():
        sorted_tile_column_names = compaction_utils.get_sorted_tile_column_names(
            agg_window, agg_window_indexes_to_agg_groups[agg_window]
        )
        array_agg_columns.append(
            F.array(*[F.col(col_name) for col_name in sorted_tile_column_names]).alias(str(agg_window))
        )

    return output_df.select(*group_by_cols, *array_agg_columns)



import contextlib
import logging
import time
import typing
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import attrs
import numpy
import pandas
import pyspark
import pytz
from pyspark.sql import dataframe as pyspark_dataframe

import tecton.framework
import tecton_core.query.dialect
from tecton._internals.offline_store_credentials import INTERACTIVE_OFFLINE_STORE_OPTIONS_PROVIDERS
from tecton._internals.rewrite import rewrite_tree_for_spine
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.secret_resolver import LocalDevSecretResolver
from tecton.framework import configs
from tecton.snowflake_context import SnowflakeContext
from tecton_athena import athena_session
from tecton_athena.data_catalog_helper import register_feature_view_as_athena_table_if_necessary
from tecton_athena.query.translate import AthenaSqlExecutor
from tecton_athena.query.translate import athena_convert
from tecton_core import conf
from tecton_core import data_types
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.pandas_compat import pandas_to_spark
from tecton_core.query.dialect import Dialect
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.node_interface import DataframeWrapper
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import recurse_query_tree
from tecton_core.query.node_utils import get_first_input_node_of_class
from tecton_core.query.node_utils import get_pipeline_dialect
from tecton_core.query.node_utils import get_staging_nodes
from tecton_core.query.node_utils import tree_contains
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.pandas import node
from tecton_core.schema import Schema
from tecton_core.time_utils import convert_pandas_df_for_snowflake_upload
from tecton_spark import schema_spark_utils
from tecton_spark.query import translate
from tecton_spark.type_annotations import PySparkDataFrame
from tecton_spark.type_annotations import PySparkSession


if typing.TYPE_CHECKING:
    import snowflake.snowpark

logger = logging.getLogger(__name__)

# We have to use Any here because snowflake.snowpark.DataFrame is not a direct dependency of the SDK.
snowpark_dataframe = Any

_internal_index_column = "_tecton_internal_index_col"


def convert_pandas_timestamps_from_spark(pandas_df) -> pandas.DataFrame:
    """Match pandas timezone to that of Spark, s.t. the timestamps are correctly displayed."""
    from tecton.tecton_context import TectonContext

    tz = TectonContext.get_instance()._spark.conf.get("spark.sql.session.timeZone")
    for col in pandas_df.columns:
        if pandas.core.dtypes.common.is_datetime64_dtype(pandas_df[col]):
            pandas_df[col] = pandas_df[col].dt.tz_localize(pytz.timezone(tz))
            pandas_df[col] = pandas_df[col].dt.tz_convert(pytz.timezone("UTC"))
            if conf.get_bool("TECTON_STRIP_TIMEZONE_FROM_FEATURE_VALUES"):
                pandas_df[col] = pandas_df[col].dt.tz_localize(None)
    # TODO: We're editing the df in place, but returning the same object. This would be less confusing if we picked
    #  one or the other
    return pandas_df


def _remove_timezones(df: pandas.DataFrame) -> pandas.DataFrame:
    for col in df.columns:
        if pandas.core.dtypes.common.is_datetime64_any_dtype(df[col]):
            df[col] = df[col].dt.tz_localize(None)
    return df


@attrs.define(auto_attribs=True)
class FeatureVector(object):
    """
    FeatureVector Class.

    A FeatureVector is a representation of a single feature vector. Usage of a FeatureVector typically involves
    extracting the feature vector using ``to_pandas()``, ``to_dict()``, or ``to_numpy()``.

    """

    _names: List[str]
    _values: List[Union[int, str, bytes, float, list]]
    _effective_times: List[Optional[datetime]]
    slo_info: Optional[Dict[str, str]] = None

    @sdk_public_method
    def to_dict(
        self, return_effective_times: bool = False
    ) -> Dict[str, Union[int, str, bytes, float, list, dict, None]]:
        """Turns vector into a Python dict.

        :param: return_effective_times: Whether to return the effective time of the feature.

        :return: A Python dict.
        """
        if return_effective_times:
            return {
                name: {"value": self._values[i], "effective_time": self._effective_times[i]}
                for i, name in enumerate(self._names)
            }

        return dict(zip(self._names, self._values))

    @sdk_public_method
    def to_pandas(self, return_effective_times: bool = False) -> pandas.DataFrame:
        """Turns vector into a Pandas DataFrame.

        :param: return_effective_times: Whether to return the effective time of the feature as part of the DataFrame.

        :return: A Pandas DataFrame.
        """
        if return_effective_times:
            return pandas.DataFrame(
                list(zip(self._names, self._values, self._effective_times)), columns=["name", "value", "effective_time"]
            )

        return pandas.DataFrame([self._values], columns=self._names)

    @sdk_public_method
    def to_numpy(self, return_effective_times: bool = False) -> numpy.array:
        """Turns vector into a numpy array.

        :param: return_effective_times: Whether to return the effective time of the feature as part of the list.

        :return: A numpy array.
        """
        if return_effective_times:
            return numpy.array([self._values, self._effective_times])

        return numpy.array(self._values)

    def _update(self, other: "FeatureVector"):
        self._names.extend(other._names)
        self._values.extend(other._values)
        self._effective_times.extend(other._effective_times)


@dataclass
class GetFeaturesForEventsParams:
    fco: Union[specs.FeatureServiceSpec, specs.FeatureViewSpec]
    events: Union[pyspark_dataframe.DataFrame, pandas.DataFrame, "TectonDataFrame", str]
    timestamp_key: Optional[str] = None
    from_source: Optional[bool] = None
    compute_mode: Optional[Union[ComputeMode, str]] = None


@dataclass
class GetFeaturesInRangeParams:
    fco: Union[specs.FeatureServiceSpec, specs.FeatureViewSpec]
    start_time: datetime
    end_time: datetime
    entities: Optional[Union[pyspark_dataframe.DataFrame, pandas.DataFrame, "TectonDataFrame"]] = None
    max_lookback: Optional[timedelta] = None
    from_source: Optional[bool] = None
    compute_mode: Optional[Union[ComputeMode, str]] = None


@attrs.define(repr=False)
class TectonDataFrame(DataframeWrapper):
    """
    A thin wrapper around Pandas, Spark, and Snowflake dataframes.
    """

    _spark_df: Optional[PySparkDataFrame] = None
    _pandas_df: Optional[pandas.DataFrame] = None
    # TODO: Change the type to snowflake.snowpark.DataFrame, currently it will
    # fail type checking for our auto generated doc.
    _snowflake_df: Optional[Any] = None
    # This should not be accessed directly. use _querytree instead.
    _querytree_do_not_use: Optional[NodeRef] = attrs.field(default=None, repr=lambda x: "TectonQueryTree")

    _need_rewrite_tree: bool = False

    # _schema is the schema of TectonDataFrame. It is present only if the TectonDataFrame is built from a pandas
    # dataframe as a spine and it is used when converting a pandas datafarme to a spark dataframe. Note the _schema only
    # contains the name and data type for those columns Tecton manages. If the spine contains extra columns like `label`
    # etc, those columns are converted to their default spark data types.
    _schema: Optional[Schema] = None

    # This is the params that are passed to get_features_for_events or get_features_in_range.
    _request_context: Optional[Union[GetFeaturesInRangeParams, GetFeaturesForEventsParams]] = None

    @property
    def _querytree(self):
        if self._need_rewrite_tree:
            rewrite_tree_for_spine(self._querytree_do_not_use)
            self._need_rewrite_tree = False
        return self._querytree_do_not_use

    @sdk_public_method
    def start_dataset_job(
        self,
        dataset_name: Optional[str] = None,
        cluster_config: Optional[configs.ComputeConfigTypes] = None,
        tecton_materialization_runtime: Optional[str] = None,
        environment: Optional[str] = None,
        extra_config: Optional[Dict[str, Any]] = None,
    ) -> "tecton.framework.dataset.Dataset":
        """
        :param dataset_name: Dataset object will be created with this name.
                Dataset can be later retrieved by this name,
                hence it must be unique within the workspace.
        :param cluster_config: Configuration for Spark/Rift cluster
        :param tecton_runtime: Version of `tecton` package used by the job cluster
        :param environment: The custom environment in which jobs will be run
        :param extra_config: Additional parameters (the list may vary depending on
                the tecton runtime) which may be used to tune remote execution heuristics
                (ie, what number to use when chunking the events dataframe)
        :return: DatasetJob object
        """
        return None

    @sdk_public_method
    def explain(
        self,
        node_id: bool = True,
        name: bool = True,
        description: bool = True,
        columns: bool = False,
    ):
        """Prints the query tree. Should only be used when this TectonDataFrame is backed by a query tree.

        Args:
            node_id: If True, the unique id associated with each node will be rendered.
            name: If True, the class names of the nodes will be rendered.
            description: If True, the actions of the nodes will be rendered.
            columns: If True, the columns of each node will be rendered as an appendix after tree itself.
        """
        if self._querytree:
            if not name and not description:
                msg = "At least one of 'name' or 'description' must be True."
                raise RuntimeError(msg)
            if columns and not node_id:
                msg = "Can only show columns if 'node_id' is True."
                raise RuntimeError(msg)
            print(self._querytree.pretty_str(node_id=node_id, name=name, description=description, columns=columns))
        else:
            print("Explain is only available for TectonDataFrames backed by a query tree.")

    @sdk_public_method
    def to_spark(self) -> PySparkDataFrame:
        """Returns data as a Spark DataFrame.

        :return: A Spark DataFrame.
        """
        if self._spark_df is not None:
            return self._spark_df
        else:
            from tecton.tecton_context import TectonContext

            tc = TectonContext.get_instance()
            if self._querytree is not None:
                return self._to_spark(tc._spark)
            elif self._pandas_df is not None:
                if self._schema is not None:
                    extra_columns = list(set(self._pandas_df.columns) - set(self._schema.column_names()))
                    if len(extra_columns) == 0:
                        pdf = self._pandas_df[self._schema.column_names()]
                        return pandas_to_spark(tc._spark, pdf, schema=schema_spark_utils.schema_to_spark(self._schema))
                    else:
                        # If there are extra columns beyond Tecton's management sopce, it splits the spine into two
                        # parts:
                        #   1. sub_df_1, which contains Tecton managed columns, is built with explicit schema.
                        #   2. sub_df_2, which contains those extra columns, is built with spark default schema(no
                        #      explicit schema passed in.
                        # Eventually these two parts are joined together using an internal index column, which is
                        # dropped afterwards. Note the join operation isn't expensive here given it is backed by a
                        # pandas dataframe that is already loaded into memory.
                        pdf = self._pandas_df.rename_axis(_internal_index_column).reset_index()
                        pdf_schema = self._schema + Schema.from_dict({_internal_index_column: data_types.Int64Type()})
                        sub_df_1 = pandas_to_spark(
                            tc._spark,
                            pdf[pdf_schema.column_names()],
                            schema=schema_spark_utils.schema_to_spark(pdf_schema),
                        )
                        sub_df_2 = pandas_to_spark(tc._spark, pdf[[_internal_index_column, *extra_columns]])
                        return sub_df_1.join(sub_df_2, on=_internal_index_column).drop(_internal_index_column)
                else:
                    return pandas_to_spark(tc._spark, self._pandas_df)
            else:
                raise NotImplementedError

    @sdk_public_method
    def to_pandas(self, pretty_sql: bool = False) -> pandas.DataFrame:
        """Returns data as a Pandas DataFrame.

        Args:
            pretty_sql: Not applicable when using spark. For Snowflake and Athena, to_pandas() will generate a SQL string, execute it, and then return the resulting data in a pandas DataFrame.
            If True, the sql will be reformatted and executed as a more readable,
            multiline string. If False, the SQL will be executed as a one line string. Use pretty_sql=False for better performance.

        :return: A Pandas DataFrame.
        """
        if self._pandas_df is not None:
            return self._pandas_df

        assert self._spark_df is not None or self._snowflake_df is not None or self._querytree is not None

        if self._spark_df is not None:
            return convert_pandas_timestamps_from_spark(self._spark_df.toPandas())

        if self._snowflake_df is not None:
            return self._snowflake_df.to_pandas(statement_params={"SF_PARTNER": "tecton-ai"})

        if self._querytree is not None:
            if self._compute_mode == ComputeMode.ATHENA:
                output = self._to_pandas_from_athena(pretty_sql)
            elif self._compute_mode == ComputeMode.RIFT:
                output = self._to_pandas_from_duckdb(pretty_sql)
                if conf.get_bool("TECTON_STRIP_TIMEZONE_FROM_FEATURE_VALUES"):
                    output = _remove_timezones(output)
            elif self._compute_mode == ComputeMode.SNOWFLAKE:
                output = self._to_pandas_from_snowflake(pretty_sql)
            elif self._compute_mode == ComputeMode.SPARK:
                output = convert_pandas_timestamps_from_spark(self.to_spark().toPandas())
            else:
                raise ValueError(self._compute_mode)
            # Cache this since it's important to keep this around. We call this method repeatedly e.g. to do spine
            # validation, infer timestamps, etc. We also generate a unique id based on this object and rely on that
            # to be stable.
            self._pandas_df = output
            return output

    @property
    def _compute_mode(self):
        assert self._querytree is not None, "Must be called on a QueryTree DataFrame"
        return self._querytree.node.compute_mode

    def _to_spark(self, spark: PySparkSession) -> PySparkDataFrame:
        return translate.spark_convert(self._querytree, spark).to_dataframe(spark)

    @staticmethod
    def _cast_timestamps_to_utc_aware_pandas(df: pandas.DataFrame) -> pandas.DataFrame:
        # Converts all datetime columns within the DF to UTC datetimes at microsecond precision
        datetime_cols = df.select_dtypes(include=["datetime"]).columns
        for col in datetime_cols:
            # Localize where necessary, because pandas refuses to cast from a tz naive object
            df[col] = df[col].dt.tz_localize("UTC")
        datetime_cols = df.select_dtypes(include=[pandas.DatetimeTZDtype]).columns
        for col in datetime_cols:
            # Now we've got tz-aware datetimes, so we can convert explicitly to UTC
            df[col] = df[col].dt.tz_convert("UTC").astype("datetime64[us, UTC]")
        # TODO: We're editing the df in place, but returning the same object. This would be less confusing if we picked
        #  one or the other
        return df

    def _to_pandas_from_duckdb(self, pretty_sql: bool) -> pandas.DataFrame:
        from tecton_core.query import query_tree_executor

        # TODO(danny): support pretty_sql
        fv_dialect = get_pipeline_dialect(self._querytree)
        if fv_dialect is not None and fv_dialect not in (
            Dialect.SNOWFLAKE,
            Dialect.PANDAS,
            Dialect.BIGQUERY,
        ):
            msg = f"Rift does not support feature views of {fv_dialect.value} dialect."
            raise Exception(msg)

        executor = query_tree_executor.QueryTreeExecutor(
            secret_resolver=LocalDevSecretResolver(),
            offline_store_options_providers=INTERACTIVE_OFFLINE_STORE_OPTIONS_PROVIDERS,
        )

        qt_output_df = executor.exec_qt(self._querytree).result_df
        if not conf.get_bool("TECTON_STRIP_TIMEZONE_FROM_FEATURE_VALUES"):
            qt_output_df = TectonDataFrame._cast_timestamps_to_utc_aware_pandas(qt_output_df)

        return qt_output_df

    def _to_pandas_from_athena(self, pretty_sql: bool) -> pandas.DataFrame:
        with self._register_tables(tecton_core.query.dialect.Dialect.ATHENA):
            session = athena_session.get_session()
            recurse_query_tree(
                self._querytree,
                lambda node: register_feature_view_as_athena_table_if_necessary(
                    node.feature_definition_wrapper, session
                )
                if isinstance(node, OfflineStoreScanNode)
                else None,
            )
            views = self._register_querytree_views_or_tables(pretty_sql=pretty_sql)
            try:
                athena_sql_executor = AthenaSqlExecutor(session)
                df = athena_convert(self._querytree, athena_sql_executor, pretty_sql).to_dataframe()
            finally:
                # A tree and its subtree can share the same temp tables. If to_pandas() is called
                # concurrently, tables can be deleted when they are still being used by the other tree.
                for view_name, _ in views:
                    session.delete_view_if_exists(view_name)
            return df

    def _to_pandas_from_snowflake(self, pretty_sql: bool) -> pandas.DataFrame:
        with self._register_tables(tecton_core.query.dialect.Dialect.SNOWFLAKE):
            return self._get_snowflake_exec_node(pretty_sql=pretty_sql).to_dataframe()

    def _get_snowflake_exec_node(self, pretty_sql: bool) -> node.SqlExecNode:
        """Helper method used to create Snowflake QT"""
        from tecton_snowflake.query.translate import SnowparkExecutor
        from tecton_snowflake.query.translate import snowflake_convert

        # TODO(TEC-15833): This requires snowpark. Decide if we will allow a snowflake connection instead of snowpark session
        snowflake_session = SnowflakeContext.get_instance().get_session()
        snowpark_executor = SnowparkExecutor(snowflake_session)
        # This rewrites the tree and needs to be called before registering temp tables/views.
        sql_exec_node = snowflake_convert(self._querytree, snowpark_executor, pretty_sql=pretty_sql)

        views = self._register_querytree_views_or_tables(pretty_sql=pretty_sql)
        return sql_exec_node

    def _to_sql(self, pretty_sql: bool = False):
        if self._querytree is not None:
            if tree_contains(self._querytree, MultiOdfvPipelineNode):
                # SQL is not available for ODFVs. Showing SQL only for the subtree below the ODFV pipeline
                subtree = self.get_sql_node(self._querytree)
                return subtree.to_sql(pretty_sql=pretty_sql)
            else:
                return self._querytree.to_sql(pretty_sql=pretty_sql)
        else:
            raise NotImplementedError

    def get_sql_node(self, tree: NodeRef):
        # Returns the first node from which SQL can be generated(subtree below ODFV pipeline)
        can_be_pushed = (
            MultiOdfvPipelineNode,
            RenameColsNode,
        )
        if isinstance(tree.node, can_be_pushed):
            return self.get_sql_node(tree.node.input_node)
        return tree

    @contextlib.contextmanager
    def _register_tables(self, dialect: tecton_core.query.dialect.Dialect):
        tables = set()

        def maybe_register_temp_table(node):
            if isinstance(node, UserSpecifiedDataNode):
                assert isinstance(node.data, TectonDataFrame), "Must be called with a TectonDataFrame"
                # Don't register the same table twice
                if node.data._temp_table_name not in tables:
                    tmp_table_name = node.data._register_temp_table(dialect)
                    tables.add(tmp_table_name)

        recurse_query_tree(
            self._querytree,
            maybe_register_temp_table,
        )

        try:
            yield
        finally:
            self._drop_temp_tables(dialect, tables)

    def _get_querytree_sql_views(self, pretty_sql=False):
        qt_views = []
        recurse_query_tree(self._querytree, lambda node: qt_views.extend(node.get_sql_views(pretty_sql=pretty_sql)))
        return qt_views

    def _register_querytree_views_or_tables(self, pretty_sql=False) -> typing.Sequence[Tuple[str, str]]:
        views = self._get_querytree_sql_views(pretty_sql=pretty_sql)
        if self._compute_mode == ComputeMode.ATHENA:
            session = athena_session.get_session()
            for view_name, view_sql in views:
                session.create_view(view_sql, view_name)
        elif self._compute_mode == ComputeMode.SNOWFLAKE:
            import tecton_snowflake.query.dataframe_helper as snowflake_dataframe_helper

            snowpark_session = SnowflakeContext.get_instance().get_session()
            snowflake_dataframe_helper.register_temp_views_or_tables(views=views, snowpark_session=snowpark_session)
        else:
            raise ValueError(self._compute_mode)
        return views

    def _to_snowpark_from_snowflake(self, pretty_sql: bool) -> snowpark_dataframe:
        with self._register_tables(tecton_core.query.dialect.Dialect.SNOWFLAKE):
            return self._get_snowflake_exec_node(pretty_sql=pretty_sql).to_snowpark()

    @sdk_public_method
    def to_snowpark(self, pretty_sql: bool = False) -> snowpark_dataframe:
        """
        Returns data as a Snowpark DataFrame.

        Args:
            pretty_sql: to_snowpark() will generate a SQL string, execute it, and then return the resulting data in a snowpark DataFrame.
            If True, the sql will be reformatted and executed as a more readable, multiline string.
            If False, the SQL will be executed as a one line string. Use pretty_sql=False for better performance.

        :return: A Snowpark DataFrame.
        """
        if self._snowflake_df is not None:
            return self._snowflake_df

        if self._querytree is not None and self._compute_mode == ComputeMode.SNOWFLAKE:
            return self._to_snowpark_from_snowflake(pretty_sql)

        assert self._pandas_df is not None
        return SnowflakeContext.get_instance().get_session().createDataFrame(self._pandas_df)

    @classmethod
    def _create(
        cls,
        df: Union[PySparkDataFrame, pandas.DataFrame, NodeRef],
        rewrite: bool = True,
    ):
        """Creates a Tecton DataFrame from a Spark or Pandas DataFrame."""
        if isinstance(df, pandas.DataFrame):
            return cls(spark_df=None, pandas_df=df, snowflake_df=None)
        elif isinstance(df, PySparkDataFrame):
            return cls(spark_df=df, pandas_df=None, snowflake_df=None)
        elif isinstance(df, NodeRef):
            return cls(
                spark_df=None, pandas_df=None, snowflake_df=None, querytree_do_not_use=df, need_rewrite_tree=rewrite
            )

        msg = f"DataFrame must be of type pandas.DataFrame or pyspark.sql.Dataframe, not {type(df)}"
        raise TypeError(msg)

    @classmethod
    def _create_from_pandas_with_schema(cls, df: pandas.DataFrame, schema: Schema):
        if isinstance(df, pandas.DataFrame):
            return cls(spark_df=None, pandas_df=df, snowflake_df=None, schema=schema)
        msg = f"DataFrame must be pandas.DataFrame when using _create_from_pandas, not {type(df)}"
        raise TypeError(msg)

    @classmethod
    # This should be merged into _create once snowpark is installed with pip
    def _create_with_snowflake(cls, df: "snowflake.snowpark.DataFrame"):
        """Creates a Tecton DataFrame from a Snowflake DataFrame."""
        from snowflake.snowpark import DataFrame as SnowflakeDataFrame

        if isinstance(df, SnowflakeDataFrame):
            return cls(spark_df=None, pandas_df=None, snowflake_df=df)

        msg = f"DataFrame must be of type snowflake.snowpark.Dataframe, not {type(df)}"
        raise TypeError(msg)

    @classmethod
    def _create_with_snowflake_sql(cls, sql: str):
        if isinstance(sql, str):
            return cls(
                spark_df=None, pandas_df=None, snowflake_df=SnowflakeContext.get_instance().get_session().sql(sql)
            )

        msg = f"sql must be of type str, not {type(sql)}"
        raise TypeError(msg)

    def subtree(self, node_id: int) -> "TectonDataFrame":
        """Creates a TectonDataFrame from a subtree of prior querytree labeled by a node id in .explain()."""
        if not self._querytree:
            msg = "Cannot construct a TectonDataFrame from a node id."
            raise RuntimeError(msg)

        tree = self._querytree.create_tree()
        qt_root = NodeRef(tree.get_node(node_id).data)

        if self._compute_mode == ComputeMode.RIFT:
            # Rift requires StagingNodes to properly execute a querytree with a DataSourceScanNode or a
            # FeatureViewPipelineNode. If a StagingNode is not present, we add one.
            data_source_staging_nodes = get_staging_nodes(qt_root, QueryTreeStep.DATA_SOURCE)
            data_source_scan_node = get_first_input_node_of_class(qt_root, node_class=DataSourceScanNode)
            pipeline_staging_nodes = get_staging_nodes(qt_root, QueryTreeStep.PIPELINE)
            feature_view_pipeline_node = get_first_input_node_of_class(qt_root, node_class=FeatureViewPipelineNode)
            if len(data_source_staging_nodes) == 0 and data_source_scan_node is not None:
                staging_node = StagingNode(
                    dialect=Dialect.DUCKDB,
                    compute_mode=ComputeMode.RIFT,
                    input_node=qt_root.deepcopy(),
                    staging_table_name=data_source_scan_node.ds.name,
                    query_tree_step=QueryTreeStep.DATA_SOURCE,
                )
                qt_root.node = staging_node
            elif len(pipeline_staging_nodes) == 0 and feature_view_pipeline_node is not None:
                staging_node = StagingNode(
                    dialect=Dialect.DUCKDB,
                    compute_mode=ComputeMode.RIFT,
                    input_node=qt_root.deepcopy(),
                    staging_table_name=feature_view_pipeline_node.feature_definition_wrapper.name,
                    query_tree_step=QueryTreeStep.PIPELINE,
                )
                qt_root.node = staging_node

        # Do not apply rewrites again as they should have already been applied when generating the query tree for this
        # TectonDataFrame.
        return TectonDataFrame._create(qt_root, rewrite=False)

    def _timed_to_pandas(self):
        """Convenience method for measuring performance."""
        start = time.time()
        ret = self.to_spark().toPandas()
        end = time.time()
        print(f"took {end - start} seconds")
        return ret

    # The methods below implement the DataframeWrapper interface
    def _register_temp_table(self, dialect: tecton_core.query.dialect.Dialect):
        if dialect == tecton_core.query.dialect.Dialect.SPARK:
            self.to_spark().createOrReplaceTempView(self._temp_table_name)
            return
        if dialect == tecton_core.query.dialect.Dialect.ATHENA:
            session = athena_session.get_session()
            session.write_pandas(self.to_pandas(), self._temp_table_name)
        elif dialect == tecton_core.query.dialect.Dialect.SNOWFLAKE:
            session = SnowflakeContext.get_instance().get_session()
            if self._snowflake_df:
                self._snowflake_df.write.mode("overwrite").save_as_table(
                    table_name=self._temp_table_name, table_type="temporary"
                )
            else:
                df_to_write = self.to_pandas()
                convert_pandas_df_for_snowflake_upload(df_to_write)
                session.write_pandas(
                    df_to_write,
                    table_name=self._temp_table_name,
                    auto_create_table=True,
                    table_type="temporary",
                    quote_identifiers=False,
                    overwrite=True,
                    parallel=64,
                    chunk_size=1000000,
                    use_logical_type=True,
                )
        elif dialect == tecton_core.query.dialect.Dialect.DUCKDB:
            from tecton_core.duckdb_context import DuckDBContext

            pandas_df = self.to_pandas()
            connection = DuckDBContext.get_instance().get_connection()
            connection.sql(f"CREATE OR REPLACE TABLE {self._temp_table_name} AS SELECT * FROM pandas_df")
        else:
            msg = f"Unexpected dialect {dialect}"
            raise Exception(msg)

        return self._temp_table_name

    def _drop_temp_tables(self, dialect: tecton_core.query.dialect.Dialect, tables: typing.Collection[str]):
        if dialect == tecton_core.query.dialect.Dialect.ATHENA:
            session = athena_session.get_session()
            for temp_table in tables:
                session.delete_table_if_exists(session.get_database(), temp_table)
        elif dialect == tecton_core.query.dialect.Dialect.SPARK:
            from tecton.tecton_context import TectonContext

            spark = TectonContext.get_instance()._spark
            for temp_table in tables:
                # DROP VIEW/DROP TABLE sql syntax is invalidated when spark_catalog is set to DeltaCatalog on EMR clusters
                if (
                    spark.conf.get("spark.sql.catalog.spark_catalog", "")
                    == "org.apache.spark.sql.delta.catalog.DeltaCatalog"
                ):
                    spark.catalog.dropTempView(temp_table)
                else:
                    spark.sql(f"DROP VIEW IF EXISTS {temp_table}")
        elif dialect == tecton_core.query.dialect.Dialect.SNOWFLAKE:
            # Snowpark automatically drops temp views when the session ends
            # All other temp views/tables are created with 'overwrite' mode
            pass

    def _select_distinct(self, columns: List[str]) -> DataframeWrapper:
        if self._pandas_df is not None:
            return TectonDataFrame._create(self._pandas_df[columns].drop_duplicates())
        elif self._spark_df is not None:
            return TectonDataFrame._create(self._spark_df.select(columns).distinct())
        else:
            raise NotImplementedError

    @property
    def _dataframe(self) -> Union[pyspark.sql.DataFrame, pandas.DataFrame]:
        if self._spark_df is not None:
            return self._spark_df
        elif self._pandas_df is not None:
            return self._pandas_df
        elif self._snowflake_df is not None:
            return self._snowflake_df
        elif self._querytree:
            if self._compute_mode == ComputeMode.SPARK:
                return self.to_spark()
            else:
                return self.to_pandas()
        else:
            raise NotImplementedError

    @property
    def columns(self) -> Sequence[str]:
        if self._querytree:
            return self._querytree.columns
        elif self._spark_df is not None:
            return self._spark_df.columns
        elif self._pandas_df is not None:
            return list(self._pandas_df.columns)
        elif self._snowflake_df is not None:
            return self._snowflake_df.columns
        else:
            raise NotImplementedError


# for legacy compat
DataFrame = TectonDataFrame



import itertools
from typing import Iterable
from typing import TypeVar


T = TypeVar("T")


def batched_iterator(iterable: Iterable[T], batch_size: int) -> Iterable[Iterable[T]]:
    it = iterable.__iter__()
    if batch_size < 1:
        msg = "Batch size must be at least 1"
        raise ValueError(msg)
    while True:
        try:
            peeked = next(it)
        except StopIteration:
            return

        def batch():
            yield peeked
            yield from itertools.islice(it, batch_size - 1)

        yield batch()



from base64 import b64encode
from typing import Any
from typing import Iterable
from typing import List
from typing import Union
from urllib.parse import urlencode

import pandas
from pyspark import sql as pyspark_sql

from tecton._internals import errors
from tecton._internals import ingest_utils
from tecton._internals import metadata_service
from tecton_core import feature_definition_wrapper
from tecton_core import id_helper
from tecton_core.conf import tecton_url
from tecton_core.errors import SCHEMA_VALIDATION_COLUMN_TYPE_MISMATCH_ERROR
from tecton_core.schema import Schema
from tecton_proto.common import fco_locator__client_pb2 as fco_locator_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2
from tecton_proto.online_store import feature_value__client_pb2 as feature_value_pb2
from tecton_spark.schema_spark_utils import schema_from_spark


KEY_DELETION_MAX = 500000


def delete_keys(
    online: bool,
    offline: bool,
    keys: Union[pyspark_sql.DataFrame, pandas.DataFrame],
    feature_definition: feature_definition_wrapper.FeatureDefinitionWrapper,
) -> List[str]:
    if not offline and not online:
        raise errors.NO_STORE_SELECTED

    materialization_state_transitions = feature_definition.fv_spec.materialization_state_transitions
    if offline and any(transition.offline_enabled for transition in materialization_state_transitions):
        if not feature_definition.offline_store_config.HasField("delta"):
            raise errors.OFFLINE_STORE_NOT_SUPPORTED

    if online and all(not transition.online_enabled for transition in materialization_state_transitions):
        print("Online materialization was never enabled. No data to be deleted in online store.")
        online = False

    if offline and all(not transition.offline_enabled for transition in materialization_state_transitions):
        print("Offline materialization was never enabled. No data to be deleted in offline store.")
        offline = False

    if not (offline or online):
        return None

    if not isinstance(keys, (pyspark_sql.DataFrame, pandas.DataFrame)):
        raise errors.INVALID_JOIN_KEY_TYPE(type(keys))
    _validate_entity_deletion_keys_dataframe(
        df=keys, join_keys=feature_definition.join_keys, view_schema=feature_definition.view_schema
    )

    keys = keys if isinstance(keys, pandas.DataFrame) else keys.toPandas()
    keys = keys.drop_duplicates()
    info_response = _get_delete_entities_info(feature_definition.id)
    s3_path = info_response.df_path
    offline_join_keys_path = s3_path + "/offline"
    online_join_keys_path = s3_path + "/online"
    # TODO(TEC-17028): Only upload entity keys once.
    if online:
        join_key_df = _serialize_join_keys(keys, feature_definition.join_keys)
        ingest_utils.upload_df_pandas(info_response.signed_url_for_df_upload_online, join_key_df, parquet=False)
    if offline:
        ingest_utils.upload_df_pandas(info_response.signed_url_for_df_upload_offline, keys)

    return _send_deletion_request(feature_definition, online, offline, online_join_keys_path, offline_join_keys_path)


def _send_deletion_request(
    feature_definition: feature_definition_wrapper.FeatureDefinitionWrapper,
    online: bool,
    offline: bool,
    online_join_keys_path: str,
    offline_join_keys_path: str,
) -> List[str]:
    deletion_request = metadata_service_pb2.DeleteEntitiesRequest(
        fco_locator=fco_locator_pb2.FcoLocator(
            id=id_helper.IdHelper.from_string(feature_definition.id), workspace=feature_definition.workspace
        ),
        online=online,
        offline=offline,
        online_join_keys_full_path=online_join_keys_path,
        offline_join_keys_path=offline_join_keys_path,
    )
    delete_entities_response = metadata_service.instance().DeleteEntities(deletion_request)
    url = f"{tecton_url()}/app/jobs?" + urlencode(
        {
            "workspaces": feature_definition.workspace,
            "task_type": "Deletion",
            "feature_views": feature_definition.name,
        }
    )
    print(f"Deletion occurs asynchronously. View the status of the completion job at {url}")
    return list(delete_entities_response.job_ids)


def _get_delete_entities_info(string_id: str) -> metadata_service_pb2.GetDeleteEntitiesInfoResponse:
    info_request = metadata_service_pb2.GetDeleteEntitiesInfoRequest(
        feature_definition_id=id_helper.IdHelper.from_string(string_id),
    )
    return metadata_service.instance().GetDeleteEntitiesInfo(info_request)


def _serialize_join_keys(join_keys_df: pandas.DataFrame, join_keys: List[str]) -> pandas.DataFrame:
    def _serialize_fn(items: Iterable[Any]) -> str:
        ret = feature_value_pb2.FeatureValueList()
        for item in items:
            if isinstance(item, int):
                ret.feature_values.add().int64_value = item
            elif isinstance(item, str):
                ret.feature_values.add().string_value = item
            elif item is None:
                ret.feature_values.add().null_value.CopyFrom(feature_value_pb2.NullValue())
            else:
                msg = f"Unknown type: {type(item)}"
                raise Exception(msg)
        return b64encode(ret.SerializeToString()).decode()

    keys_df = join_keys_df[join_keys]
    return keys_df.apply(lambda row: _serialize_fn(tuple(row)), axis=1).to_frame(name="join_keys_array")


def _validate_entity_deletion_keys_dataframe(
    df: Union[pyspark_sql.DataFrame, pandas.DataFrame], join_keys: List[str], view_schema: Schema
) -> None:
    if len(set(df.columns)) != len(df.columns):
        raise errors.DUPLICATED_COLS_IN_KEYS(", ".join(list(df.columns)))

    row_count = df.count() if isinstance(df, pyspark_sql.DataFrame) else len(df)
    if row_count > KEY_DELETION_MAX:
        raise errors.TOO_MANY_KEYS(KEY_DELETION_MAX)
    if row_count == 0:
        msg = "keys"
        raise errors.EMPTY_ARGUMENT(msg)

    if set(df.columns) != set(join_keys):
        raise errors.INCORRECT_KEYS(", ".join(list(df.columns)), ", ".join(join_keys))

    # We skip schema validation for Pandas DataFrames since there is no reliable way to map Pandas types to Tecton types yet.
    if isinstance(df, pyspark_sql.DataFrame):
        df_columns = schema_from_spark(df.schema).column_name_and_data_types()
        fv_columns = view_schema.column_name_and_data_types()
        for df_column in df_columns:
            if df_column not in fv_columns:
                raise SCHEMA_VALIDATION_COLUMN_TYPE_MISMATCH_ERROR(
                    df_column[0], next(x for x in fv_columns if x[0] == df_column[0])[1], df_column[1]
                )



import io
import os
import platform
import sys
import tarfile
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import click
import requests
from google.protobuf import json_format
from yaspin.spinners import Spinners

import tecton
from tecton import tecton_context
from tecton._internals import metadata_service
from tecton._internals.analytics import StateUpdateEventMetrics
from tecton._internals.analytics import StateUpdateResult
from tecton._internals.utils import plural
from tecton.cli import cli_utils
from tecton.cli import printer
from tecton.cli import repo_utils
from tecton.cli.command import TectonCommand
from tecton.cli.engine_renderer import PlanRenderingClient
from tecton.cli.error_utils import format_server_errors
from tecton.cli.repo_config import DEFAULT_REPO_CONFIG_NAME
from tecton.cli.test import run_tests
from tecton.cli.workspace_utils import WorkspaceType
from tecton.framework import base_tecton_object
from tecton_core import http
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonNotFoundError
from tecton_core.id_helper import IdHelper
from tecton_proto.args import fco_args__client_pb2 as fco_args_pb2
from tecton_proto.args import repo_metadata__client_pb2 as repo_metadata_pb2
from tecton_proto.data import state_update__client_pb2 as state_update_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


class EngineCommand(TectonCommand):
    def __init__(
        self,
        *args,
        apply: bool,
        upgrade_all: bool = False,
        destroy: bool = False,
        allows_suppress_recreates: bool = False,
        has_plan_id: bool = False,
        **kwargs,
    ):
        def callback(
            yes,
            json_out,
            suppress_warnings,
            workspace,  # Not used but it needs to be here to match params list.
            config: Optional[Path],
            suppress_recreates=False,
            plan_id=None,
            skip_tests=None,
        ):
            args = EngineArgs(
                skip_tests=skip_tests,
                json_out=json_out,
                no_safety_checks=yes,
                suppress_warnings=suppress_warnings,
                repo_config_path=config,
            )

            assert not (plan_id and suppress_recreates), (
                "The flag --suppress-recreates is only used when computing a new plan. If the plan passed "
                "in using --plan-id was already computed using --suppress-recreates, that behavior persists "
                "as part of the plan."
            )

            if plan_id:
                args.plan_id = plan_id
            if suppress_recreates:
                args.suppress_recreates = suppress_recreates

            return run_engine(args, apply=apply, upgrade_all=upgrade_all, destroy=destroy)

        params = [
            click.Option(
                ["--yes", "-y"],
                is_flag=True,
                default=False,
                help="Skip all confirmation prompts.",
            ),
            click.Option(
                ["--json-out"],
                default="",
                help="Output the tecton state update diff (as JSON) to the file path provided.",
            ),
            click.Option(
                ["--suppress-warnings"],
                is_flag=True,
                default=False,
                help="Disable tecton plan linting warnings.",
            ),
            click.Option(
                ["--workspace"],
                default=None,
                type=WorkspaceType(),
                help="Name of the target workspace that tecton state update request applies to.",
            ),
            click.Option(
                ["--config"],
                help="Path to the repo config yaml file. Defaults to the repo.yaml file at the Tecton repo root.",
                default=None,
                type=click.Path(exists=True, dir_okay=False, path_type=Path, readable=True),
            ),
        ]

        if has_plan_id:
            params.append(
                click.Option(["--plan-id"], default=None, type=str, help="Id of a previously computed plan to apply.")
            )
        if not destroy:
            params.append(
                click.Option(
                    ["--skip-tests/--no-skip-tests"],
                    default=False,
                    help="Disable running tests.",
                )
            )
        if allows_suppress_recreates:
            params.append(
                click.Option(
                    ["--suppress-recreates"],
                    is_flag=True,
                    default=False,
                    help="Force suppression of all recreates into in-place updates.",
                ),
            )

        super().__init__(*args, callback=callback, params=params, uses_workspace=True, **kwargs)


# TODO: This class was created to match the old style arg parse struct when we were migrating to click to avoid having
# do a deep refactoring of the code that depends on it. It should be replaced.
@dataclass
class EngineArgs:
    skip_tests: bool
    no_safety_checks: bool
    json_out: str
    suppress_warnings: bool
    repo_config_path: Optional[Path]


def update_tecton_state(
    objects: List[base_tecton_object.BaseTectonObject],
    repo_files: List[Path],
    repo_config: Optional[Path],
    repo_root: Optional[str],
    apply,
    interactive,
    upgrade_all: bool,
    workspace_name: str,
    suppress_warnings: bool = False,
    suppress_recreates: bool = False,
    json_out_path: Optional[Path] = None,
    timeout_seconds=90 * 60,
    plan_id: Optional[str] = None,
    no_color: bool = False,
) -> StateUpdateResult:
    json_out = json_out_path is not None

    if apply and plan_id:
        # Applying an existing plan, so skip preparing args.
        state_id = IdHelper.from_string(plan_id)
        query_state_update_request = metadata_service_pb2.QueryStateUpdateRequestV2(
            state_id=state_id,
            workspace=workspace_name,
            no_color=no_color,
            json_output=json_out,
            suppress_warnings=suppress_warnings,
        )

        try:
            query_state_update_response = metadata_service.instance().QueryStateUpdateV2(query_state_update_request)
        except (
            TectonInternalError,
            TectonAPIValidationError,
        ) as e:
            printer.safe_print(e)
            return StateUpdateResult.from_error_message(str(e), suppress_recreates)

        if query_state_update_response.error:
            printer.safe_print(query_state_update_response.error)
            return StateUpdateResult.from_error_message(query_state_update_response.error, suppress_recreates)
        if len(query_state_update_response.validation_errors.errors) > 0:
            # Cannot pretty-print validation result using format_server_errors(), because collected local objects
            # might have changed since this plan was generated, so can't accurately match with this plan's FCOs.
            message = "Cannot apply plan because it had errors."
            printer.safe_print(message)
            return StateUpdateResult.from_error_message(message, suppress_recreates)

    else:
        with printer.safe_yaspin(Spinners.earth, text="Collecting local feature declarations") as sp:
            fco_args, repo_source_info = _get_declared_fco_args(objects)
            sp.ok(printer.safe_string("✅"))

        new_state_update_request = metadata_service_pb2.NewStateUpdateRequestV2(
            request=state_update_pb2.StateUpdateRequest(
                workspace=workspace_name,
                upgrade_all=upgrade_all,
                sdk_version=tecton.version.get_semantic_version(),
                fco_args=fco_args,
                repo_source_info=repo_source_info,
                suppress_recreates=suppress_recreates,
            ),
            no_color=no_color,
            json_output=json_out,
            suppress_warnings=suppress_warnings,
        )

        server_side_msg_prefix = "Performing server-side feature validation"
        with printer.safe_yaspin(Spinners.earth, text=f"{server_side_msg_prefix}: Initializing.") as sp:
            try:
                new_state_update_response = metadata_service.instance().NewStateUpdateV2(new_state_update_request)

                if new_state_update_response.HasField("signed_url_for_repo_upload"):
                    _upload_files(
                        repo_files, repo_config, repo_root, new_state_update_response.signed_url_for_repo_upload
                    )
                if new_state_update_response.HasField("eager_response"):
                    query_state_update_response = new_state_update_response.eager_response
                else:
                    seconds_slept = 0
                    query_state_update_request = metadata_service_pb2.QueryStateUpdateRequestV2(
                        state_id=new_state_update_response.state_id,
                        workspace=workspace_name,
                        no_color=no_color,
                        json_output=json_out,
                        suppress_warnings=suppress_warnings,
                    )
                    while True:
                        query_state_update_response = metadata_service.instance().QueryStateUpdateV2(
                            query_state_update_request
                        )
                        if query_state_update_response.latest_status_message:
                            sp.text = f"{server_side_msg_prefix}: {query_state_update_response.latest_status_message}"
                        if query_state_update_response.ready:
                            break
                        seconds_to_sleep = 5
                        time.sleep(seconds_to_sleep)
                        seconds_slept += seconds_to_sleep
                        if seconds_slept > timeout_seconds:
                            sp.fail(printer.safe_string("⛔"))
                            printer.safe_print("Validation timed out.")
                            return StateUpdateResult.from_error_message("Validation timed out.", suppress_recreates)

                if query_state_update_response.error:
                    sp.fail(printer.safe_string("⛔"))
                    printer.safe_print(query_state_update_response.error)
                    return StateUpdateResult.from_error_message(query_state_update_response.error, suppress_recreates)
                validation_errors = query_state_update_response.validation_errors.errors
                if len(validation_errors) > 0:
                    sp.fail(printer.safe_string("⛔"))
                    format_server_errors(validation_errors, objects, repo_root)
                    return StateUpdateResult.from_error_message(str(validation_errors), suppress_recreates)
                sp.ok(printer.safe_string("✅"))
            except (TectonInternalError, TectonAPIValidationError, TectonNotFoundError) as e:
                sp.fail(printer.safe_string("⛔"))
                printer.safe_print(e)
                return StateUpdateResult.from_error_message(str(e), suppress_recreates)

        state_id = new_state_update_response.state_id

    plan_rendering_client = PlanRenderingClient(query_state_update_response)

    if not plan_rendering_client.has_diffs():
        plan_rendering_client.print_empty_plan()
    else:
        plan_rendering_client.print_plan()

        if apply:
            plan_rendering_client.print_apply_warnings()
            if interactive:
                cli_utils.confirm_or_exit(f'Are you sure you want to apply this plan to: "{workspace_name}"?')

            apply_request = metadata_service_pb2.ApplyStateUpdateRequest(state_id=state_id)
            metadata_service.instance().ApplyStateUpdate(apply_request)

            num_fcos = plan_rendering_client.num_fcos_changed
            printer.safe_print(
                f'🎉 Done! Applied changes to {num_fcos} {plural(num_fcos, "object", "objects")} in workspace "{workspace_name}".'
            )

    if json_out_path:
        repo_diff_summary = plan_rendering_client.get_json_plan_output()
        json_out_path.parent.mkdir(parents=True, exist_ok=True)
        json_out_path.write_text(repo_diff_summary)

    return StateUpdateResult(
        state_update_event_metrics=StateUpdateEventMetrics(
            num_total_fcos=len(objects),
            suppress_recreates=suppress_recreates,
            json_out=(json_out_path is not None),
            error_message=None,
            num_fcos_changed=query_state_update_response.successful_plan_output.num_fcos_changed,
            num_warnings=query_state_update_response.successful_plan_output.num_warnings,
        ),
        success_response=query_state_update_response,
    )


def run_engine(args: EngineArgs, apply: bool = False, destroy=False, upgrade_all=False) -> StateUpdateResult:
    cli_utils.check_version()

    # Resolve the json_out_filename prior to running `get_tecton_objects(...)` so
    # that relative directories in the file name are supported (`get_tecton_objects`
    # changes the working directory).
    json_out_path = None
    if args.json_out:
        json_out_path = Path(args.json_out).resolve()

    # Must use hasattr instead of args.plan_id, because only `apply` has the plan_id arg, but this
    # code path is also used by `plan`, `destroy`, and `upgrade`, which will fail on args.plan_id
    plan_id = None
    if hasattr(args, "plan_id"):
        plan_id = args.plan_id
    suppress_recreates = False
    if hasattr(args, "suppress_recreates") and args.suppress_recreates:
        suppress_recreates = True

    if destroy or plan_id:
        # There is no need to run tests when destroying a repo or when a plan_id is provided.
        top_level_objects: List[base_tecton_object.BaseTectonObject] = []
        repo_root = None
        repo_files: List[Path] = []
        repo_config = None
    else:
        top_level_objects, repo_root, repo_files, repo_config = repo_utils.get_tecton_objects_skip_validation(
            args.repo_config_path
        )

        if args.skip_tests == False:
            run_tests(args.repo_config_path)

    # When using server-side plan rendering, use no colors on Windows
    # or if NO_COLOR is set
    no_color = platform.system() == "Windows" or cli_utils.no_color_convention()

    return update_tecton_state(
        objects=top_level_objects,
        apply=apply,
        interactive=not args.no_safety_checks,
        repo_files=repo_files,
        repo_config=repo_config,
        repo_root=repo_root,
        upgrade_all=upgrade_all,
        suppress_warnings=args.suppress_warnings,
        suppress_recreates=suppress_recreates,
        json_out_path=json_out_path,
        plan_id=plan_id,
        workspace_name=tecton_context.get_current_workspace(),
        no_color=no_color,
    )


def dump_local_state(objects: Sequence[base_tecton_object.BaseTectonObject]):
    fco_args, repo_source_info = _get_declared_fco_args(objects)
    request_plan = metadata_service_pb2.NewStateUpdateRequest(
        request=state_update_pb2.StateUpdateRequest(fco_args=fco_args, repo_source_info=repo_source_info)
    )
    printer.safe_print(json_format.MessageToJson(request_plan, including_default_value_fields=True))


# upload tar.gz of python files to url via PUT request
def _upload_files(repo_files: List[Path], repo_config: Optional[Path], repo_root, url: str):
    tar_bytes = io.BytesIO()
    with tarfile.open(fileobj=tar_bytes, mode="w|gz") as targz:
        for f in repo_files:
            targz.add(f, arcname=os.path.relpath(f, repo_root))
        if repo_config:
            # Always upload the repo config to the default repo path. We use the default path because (a) the repo
            # config may not be under the repo root and (b) using the default path simplifies restoring and re-applying.
            targz.add(repo_config, arcname=os.path.relpath(DEFAULT_REPO_CONFIG_NAME, repo_root))
    for _ in range(3):
        try:
            r = http.session().put(url, data=tar_bytes.getbuffer())
            if r.status_code != 200:
                # We will get 403 (forbidden) when the signed url expires.
                if r.status_code == 403:
                    printer.safe_print(
                        "\nUploading feature repo failed due to expired session. Please retry the command."
                    )
                else:
                    printer.safe_print(f"\nUploading feature repo failed with reason: {r.reason}")
                sys.exit(1)
            return
        except requests.RequestException as e:
            last_error = e
    raise SystemExit(last_error)


def _get_declared_fco_args(
    objects: Sequence[base_tecton_object.BaseTectonObject],
) -> Tuple[List[fco_args_pb2.FcoArgs], repo_metadata_pb2.FeatureRepoSourceInfo]:
    all_args = []
    repo_source_info = repo_metadata_pb2.FeatureRepoSourceInfo()

    for fco_obj in objects:
        all_args.append(fco_obj._build_args())
        repo_source_info.source_info.append(fco_obj._source_info)

    return all_args, repo_source_info


apply = EngineCommand(
    name="apply",
    apply=True,
    allows_suppress_recreates=True,
    has_plan_id=True,
    help="Compare local feature definitions with remote state and *apply* local changes remotely.",
)

upgrade = EngineCommand(
    name="upgrade",
    apply=True,
    upgrade_all=True,
    help="Upgrade remote feature definitions.",
    hidden=True,
)

destroy = EngineCommand(
    name="destroy",
    destroy=True,
    apply=True,
    help="Destroy all registered objects in this workspace.",
)
